/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\AffineNormalizer.cpp
 *
 * @brief	Implements the affine normalizer class.
 */

#include "AffineNormalizer.h"
#include "GaussianBlur.h"
#include "Params.h"

/**
 * @fn	CAffineNormalizer::CAffineNormalizer()
 *
 * @brief	Default constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CAffineNormalizer::CAffineNormalizer()
{
}

/**
 * @fn	void CAffineNormalizer::Normalize(CMSER &mser, CGrayscaleImage &input, CGrayscaleImage &inputForNormalization, CGrayscaleImage &output)
 *
 * @brief	Performs the affine normalization.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]	mser				 	The MSER that will be used as input to perform normalization.
 * @param [in]	input				 	The input image from which normalization parameters will be estimated. This should be empty everywhere except in the MSER pixels.
 * @param [in]	inputToSample			The input iamge that is used to resample after normalization. It should be the full image for texture descriptors and the mask for shapes.
 * @param [out]	output				 	The affine normalized patch.
 */

void CAffineNormalizer::Normalize(CMSER &mser, CGrayscaleImage &input, CGrayscaleImage &inputToSample, CGrayscaleImage &output, CGrayscaleImage **rotMatrix, std::pair<double, double> &translation)
{
	// Blur the image with sigma_i
	float b = mser.GetMinorEllipseAxisLength();
	float sigmai = b*CParams::Instance()->s/CParams::Instance()->Ns;
	CGaussianBlur gb = CGaussianBlur();
	// Calculate the size of the Gaussian kernel
	int sz = (int) ceil(6*sigmai);
	// Apply blur
	CGrayscaleImage *copy;
	bool copied = false;
	// Optimization: If the Gaussian has support not bigger than 1, don't copy the inputToSample image, but just its pointer (I can do this because it won't be blurred)
	if (sz > 1)
	{
		gb.ApplyOptimizedGaussianBlur(input, &input, sz, sigmai);
		copy = new CGrayscaleImage(&inputToSample);
		copied = true;
		gb.ApplyOptimizedGaussianBlur(*copy, copy, sz, sigmai);
	}
	else
		copy = &inputToSample;


	// Perform Affine Normalization
	// First compute mask centroid and Rotation Matrix
	double cX, cY; // mask centroid
	CGrayscaleImage *A = ComputeRotationMatrix(input, cX, cY); // rotation matrix
	*rotMatrix = A;
	translation.first = cX; translation.second = cY;
	// Compute the gap to convert from virtual size to image size
	double gap = 2.0 / CParams::Instance()->Ns; // Gap to translate from virtual size (2 \times 2) to real size (Ns \times Ns)
	double x=-1, y=-1;
	// Now resample the MSER to the patch
	for (int xx=0;xx<CParams::Instance()->Ns;xx++)
	{
		y = -1;
		for (int yy=0;yy<CParams::Instance()->Ns;yy++)
		{
			// Get the resampled position
			double xFinal = x * A->GetPixel(0,0) +  y * A->GetPixel(1,0);
			xFinal = xFinal * CParams::Instance()->s + cX;
			double yFinal = x * A->GetPixel(0,1) +  y * A->GetPixel(1,1);
			yFinal = yFinal * CParams::Instance()->s + cY;
			// Set the output pixel by bilinear interpolation
			output.SetPixel(xx, yy, copy->GetBilinearInterpolatedPixel(xFinal, yFinal));
			y+=gap;
		}
		x+=gap;
	}
	//delete A;
	if (copied)
		delete copy;

}

/**
 * @fn	CAffineNormalizer::~CAffineNormalizer(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CAffineNormalizer::~CAffineNormalizer(void)
{
}

/**
 * @fn	CGrayscaleImage *CAffineNormalizer::ComputeRotationMatrix(CGrayscaleImage &mask, double &muX, double &muY)
 *
 * @brief	Calculates the rotation matrix.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in,out]	mask	The mask.
 * @param [in,out]	muX 	The mu x coordinate.
 * @param [in,out]	muY 	The mu y coordinate.
 *
 * @return	null if it fails, else the calculated rotation matrix.
 */

CGrayscaleImage *CAffineNormalizer::ComputeRotationMatrix(CGrayscaleImage &mask, double &muX, double &muY)
{
	// Calculate the mask total which I will be using all along
	// Also calculate the mask centroid
	double maskTotal = 0;
	muX = 0;
	muY = 0;
	for (int x=0;x<mask.Width();x++)
	{
		for (int y=0;y<mask.Height();y++)
		{
			maskTotal += mask.GetPixel(x,y);
			muX += x * mask.GetPixel(x,y);
			muY += y * mask.GetPixel(x,y);
		}
	}
	// Mask centroid
	muX /= maskTotal;
	muY /= maskTotal;
	// Now compute the covariance matrix (Second order statistics)
	double c11=0, c12=0,c22=0;
	for (int x=0;x<mask.Width();x++)
	{
		for (int y=0;y<mask.Height();y++)
		{
			c11 += (x-muX)*(x-muX) * mask.GetPixel(x,y);
			c12 += ((double) x-muX) * ((double) y-muY) * mask.GetPixel(x,y);
			c22 += (y-muY)*(y-muY) * mask.GetPixel(x,y);
		}
	}
	c11 /= maskTotal;
	c12 /= maskTotal;
	c22 /= maskTotal;
	// Calculate eigenvalues
	double aMinusd2 = (c11-c22)*(c11-c22);
	double b2 = c12*c12;
	double root = sqrt(aMinusd2 + 4 * b2);
	double lambda1 = (c11 + c22 + root) / 2.0;
	double lambda2 = (c11 + c22 - root) / 2.0;
	// Calculate Eigenvectors
	double ev11, ev12,ev21,ev22;
	if (fabs(c12) > 0.001)
	{
		double root1 = sqrt(b2 + ((lambda1 - c11)*(lambda1 - c11)));
		double root2 = sqrt(b2 + ((lambda2 - c11)*(lambda2 - c11)));
		ev11 = c12 / root1;
		ev21 = (lambda1 - c11) / root1;

		ev12 = c12 / root2;
		ev22 = (lambda2 - c11) / root2;
	}
	else
	{
		ev11 = 1;
		ev21 = 0;
		ev12 = 0;
		ev22 = 1;
	}
	// Now calculate A = 2RD^(1/2)
	lambda1 = 2 * sqrt(lambda1);
	lambda2 = 2 * sqrt(lambda2);

	CGrayscaleImage *A = new CGrayscaleImage(2,2);
	A->SetPixel(0,0, lambda1 * ev11);
	A->SetPixel(1,0, lambda2 * ev12);
	A->SetPixel(0,1, lambda1 * ev21);
	A->SetPixel(1,1, lambda2 * ev22);
	return A;

}
