/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\CMSER.cpp
 *
 * @brief	Implements the CMSER class. This class is responsible for storing MSER information useful for the algorithm.
 */

#include "CMSER.h"

/**
 * @fn	void CMSER::InitializeMSER(node *a, int imageWidth)
 *
 * @brief	Initializes the MSER.
 *
 * @author	Tato
 * @date	19/03/2014
 *
 * @param [in]	a 		The root node of the MSER.
 * @param	imageWidth	Width of the image.
 */

void CMSER::InitializeMSER(node *a, int imageWidth)
{
	// First process the node to obtain its pixels.
	pixelSet set = a->list;
	// Convert the MSER pixel representation of Oyallon to our pixel representation and store it.
	for (int i=0;i<set.size;i++)
	{
		int p=(set.data)[i];
		int x = p%imageWidth;
		int y = p/imageWidth;
		CPixel *pixel = new CPixel(x,y);
		_pixels->at(_pixelCount++) = pixel;
	}
	// Now, do the recursion to keep adding all pixels.
	list_node* sons=a->sons;
	while (sons!=NULL)
	{
		InitializeMSER(sons->current_node,imageWidth);
		sons=sons->next_node;
	}
}

/**
 * @fn	void CMSER::InitializeMinorAxisLenght(MSER *m)
 *
 * @brief	Initializes the minor axis lenght.
 *
 * @author	Tato
 * @date	19/03/2014
 *
 * @param [in]	m	The input MSER
 */

void CMSER::InitializeMinorAxisLenght(MSER *m)
{
	// Minor axis lenght of the fitting ellipse is the square root of the second (smallest) eigenvalue of the covariance matrix.
	_minorEllipseAxisLength = sqrtf(m->eigen2);
}

/**
 * @fn	void CMSER::InitializeCentroid(MSER *m)
 *
 * @brief	Initializes the centroid.
 *
 * @author	Tato
 * @date	19/03/2014
 *
 * @param [in]	m	The MSER.
 */

void CMSER::InitializeCentroid(MSER *m)
{
	double x=(m->sum_x/m->area);
	double y=(m->sum_y/m->area);
	_centroid.first = x;
	_centroid.second = y;
}

	/**
	 * @fn	CMSER::CMSER(MSER *mser, int imageWidth, int imageHeight, CGrayscaleImage *image, int scale, int id)
	 *
	 * @brief	Constructor.
	 *
	 * @author	Tato
	 * @date	23/02/2014
	 *
	 * @param [in]	mser 		Oyallon's MSER.
	 * @param	imageWidth   	Width of the image.
	 * @param	imageHeight  	Height of the image.
	 * @param [in]	image		Input image.
	 * @param	scale		 	Scale where the MSER has been obtained (from 0 to qtyScales-1).
	 * @param	id			 	The MSER identifier.
	 */

	CMSER::CMSER(MSER *mser, int imageWidth, int imageHeight, CGrayscaleImage *image, int scale, int id)
	{
		_imgWidth = imageWidth;
		_imgHeight = imageHeight;
		_image = image;
		_scale = scale;
		_pixels = new vector<CPixel *>((unsigned int) mser->area);
		_pixelCount = 0;
		// Transfer pixels from Oyallon's MSER into our class.
		InitializeMSER(mser->currentNode, imageWidth);
		_area = mser->area;
		_id = id;
		// Compute the centroid
		InitializeCentroid(mser);
		// Compute the Minor Axis Length of the fitting ellipse.
		InitializeMinorAxisLenght(mser);

	}

	/**
	 * @fn	CMSER::~CMSER()
	 *
	 * @brief	Destructor.
	 *
	 * @author	Tato
	 * @date	19/03/2014
	 */

	CMSER::~CMSER()
	{
		// Free all the pixels and the vector that holds them
		for (unsigned int i=0;i<_pixels->size();i++)
		{
			CPixel *p = _pixels->at(i);
			delete p;
		}
		delete _pixels;
	}
