#pragma once
#include "GrayscaleImage.h"
#include "HelperMethods.h"
#include <vector>
#include <math.h>
#include "MSER/mser.h"

/**
 * @class	CMSER
 *
 * @brief	A cmser.
 *
 * @author	Tato
 * @date	23/02/2014
 */

using namespace std;

class CMSER
{
private:
	vector<CPixel *> *_pixels;
	CDPixel _centroid;
	float _area;
	float _minorEllipseAxisLength;
	int _scale;
	int _imgWidth;
	int _imgHeight;
	int _pixelCount;
	CGrayscaleImage *_image;
	int _id;

	void InitializeMSER(node *a, int imageWidth);

	void InitializeMinorAxisLenght(MSER *m);

	void InitializeCentroid(MSER *m);
public:


	CMSER(MSER *mser, int imageWidth, int imageHeight, CGrayscaleImage *image, int scale, int id);

	~CMSER();
	inline float GetMinorEllipseAxisLength() { return _minorEllipseAxisLength; }
	inline float GetArea() { return _area; }
	inline const CDPixel *GetCentroid()	{ return &_centroid; }
	inline int GetScale() { return _scale; }
	inline int GetId() { return _id; }
	inline void SetId(int value) { _id = value; }
	inline const vector<CPixel *> *GetPixels() { return _pixels; }
	inline float GetScaledArea() { return _area * (float) pow(4.0, _scale); }
	inline CGrayscaleImage *GetImage() { return _image; }

	/**
	 * @fn	inline float CMSER::GetScaledMinorEllipseAxisLength()
	 *
	 * @brief	Gets scaled minor ellipse axis length.
	 *
	 * @author	Tato
	 * @date	23/02/2014
	 *
	 * @return	The scaled minor ellipse axis length.
	 */

	inline float GetScaledMinorEllipseAxisLength() { return (float) pow(2.0,_scale) * _minorEllipseAxisLength; }

	/**
	 * @fn	inline CGrayscaleImage* CMSER::GetBinaryMaskImage()
	 *
	 * @brief	Gets binary mask image.
	 *
	 * @author	Tato
	 * @date	23/02/2014
	 *
	 * @return	null if it fails, else the binary mask image.
	 */

	inline CGrayscaleImage* GetBinaryMaskImage()
	{
		CGrayscaleImage *res = new CGrayscaleImage(_imgWidth, _imgHeight);
		for(unsigned int i=0;i<_pixels->size();i++)
		{
			res->SetPixel((int) _pixels->at(i)->first, (int) _pixels->at(i)->second, 1.0);
		}
		return res;
	}

	/**
	 * @fn	inline CGrayscaleImage* CMSER::GetTextureImageBasedOnMask(CGrayscaleImage &img)
	 *
	 * @brief	Gets texture image based on mask.
	 *
	 * @author	Tato
	 * @date	23/02/2014
	 *
	 * @param [in,out]	img	The image.
	 *
	 * @return	null if it fails, else the texture image based on mask.
	 */

	inline CGrayscaleImage* GetTextureImageBasedOnMask()
	{
		CGrayscaleImage *res = new CGrayscaleImage(_imgWidth, _imgHeight);
		for(unsigned int i=0;i<_pixels->size();i++)
		{
			int x = (int) _pixels->at(i)->first;
			int y = (int) _pixels->at(i)->second;
			res->SetPixel(x, y, _image->GetPixel(x,y));
		}
		return res;
	}

	inline bool HasSameCentroid(CMSER &m2, bool translateToFinestGrid=false)
	{
		// The centroid distance should be smaller than 4 pixels in the finer grid
		if (!translateToFinestGrid)
		{
			// Compare two MSERs from immediate different scales (as in the creation of the MultiscaleMSERs)
			CDPixel centroidCoarseInFine = CDPixel(m2.GetCentroid()->first * 2, m2.GetCentroid()->second * 2);
			return (distanceBetweenPixels(_centroid, centroidCoarseInFine) < 4.0);
		}
		else
		{
			// Compare two MSERs with no precondition a priori (so translate into the finest grid for a fair comparison)
			double power = pow(2.0,_scale);
			CDPixel centroid = CDPixel((_centroid.first * power), (_centroid.second * power));
			power = pow(2.0,m2.GetScale());
			CDPixel centroid2 = CDPixel((m2.GetCentroid()->first * power), (m2.GetCentroid()->second * power));
			return (distanceBetweenPixels(centroid, centroid2) < 4.0);
		}
	}
	inline bool HasSimilarArea(CMSER &mserCoarse)
	{
		float a1 = GetScaledArea();
		float a2 = mserCoarse.GetScaledArea();
		float t1 = fabs(a1-a2);
		float t2 = MAX(a1,a2);
		float t3 = t1/t2;
		return (t3 < 0.2);
	}


};
