/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\GaussianBlur.cpp
 *
 * @brief	Implements the Gaussian Blur (1D or 2D).
 */

#include "GaussianBlur.h"

/**
 * @fn	void CGaussianBlur::ApplyGaussianBlur(CGrayscaleImage &img, CGrayscaleImage *res, int tap, double sigma)
 *
 * @brief	Applies the Gaussian blur.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]	img		The input image.
 * @param [out]	res		The input image with a Gaussian blur of standard deviation $sigma$ with a kernel of support $tap$
 * @param	tap		   	The support of the kernel.
 * @param	sigma	   	Standard deviation of the Gaussian.
 */

void CGaussianBlur::ApplyGaussianBlur(CGrayscaleImage &img, CGrayscaleImage *res, int tap, double sigma)
{
	// Build a kernel of size $tap \times tap$
	CKernel gaussianKernel = CKernel(tap,tap);
	// Load it with Gaussian values
	gaussianKernel.InitializeGaussian2DKernel(sigma);
	// Convolve the kernel with the image
	img.Convolution(gaussianKernel, res);
}

/**
 * @fn	void CGaussianBlur::ApplyOptimizedGaussianBlur(CGrayscaleImage &img, CGrayscaleImage *res, int tap, double sigma)
 *
 * @brief	Applies two 1D Gaussian blurs (possible because Gaussian kernel is separable)
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]		img	The input image.
 * @param [out]		res	The result.
 * @param	tap		   	The tap (size) of the Kernel.
 * @param	sigma	   	The sigma of the Gaussian.
 */

void CGaussianBlur::ApplyOptimizedGaussianBlur(CGrayscaleImage &img, CGrayscaleImage *res, int tap, double sigma)
{
	// Initialize a vertical Gaussian kernel
	CKernel gaussianKernelV = CKernel(1,tap);
	gaussianKernelV.InitializeGaussian1DKernel(sigma);
	// Initialize an horizontal Gaussian kernel
	CKernel gaussianKernelH = CKernel(tap,1);
	gaussianKernelH.InitializeGaussian1DKernel(sigma);
	// Now convolve the image with the vertical kernel
	img.Convolution(gaussianKernelV, res);
	// After, convolve the result with the horizontal kernel
	res->Convolution(gaussianKernelH, res);

}

/**
 * @fn	CGaussianBlur::CGaussianBlur(void)
 *
 * @brief	Default constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CGaussianBlur::CGaussianBlur(void)
{
}

/**
 * @fn	CGaussianBlur::~CGaussianBlur(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CGaussianBlur::~CGaussianBlur(void)
{
}
