/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\GrayscaleImage.cpp
 *
 * @brief	Implements the grayscale image class. Responsible for storing and handling grayscale images.
 */

#include "GrayscaleImage.h"
#include "IO_PNG/io_png.h"
#include <iostream>
#include <stdlib.h>
#include "HelperMethods.h"
#include <string.h>

using namespace std;

/**
 * @def	GETPIXEL(a,b) (a < (int) _width ? (b < (int) _height ? (a >= 0 ? (b >= 0 ? GetPixel(a,
 * 		b):0):0):0):0)
 *
 * @brief	A macro that reads a pixel from an image if it is in the image's bounds, otherwise return zero.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param	a	horizontal position of the pixel.
 * @param	b	vertical position of the pixel.
 */

#define GETPIXEL(a,b) (a < (int) _width ? (b < (int) _height ? (a >= 0 ? (b >= 0 ? GetPixel(a,b):0):0):0):0)

/**
 * @fn	CGrayscaleImage::CGrayscaleImage(size_t width, size_t height)
 *
 * @brief	Constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param	width 	The width.
 * @param	height	The height.
 */

CGrayscaleImage::CGrayscaleImage(size_t width, size_t height)
{
	// Allocate memory for the image.
	_img = (double *) calloc(width*height, sizeof(double));
	_width = width;
	_height = height;
}

/**
 * @fn	CGrayscaleImage::CGrayscaleImage(CGrayscaleImage *image)
 *
 * @brief	Copy constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]	image		The image being copied.
 */

CGrayscaleImage::CGrayscaleImage(CGrayscaleImage *image)
{
	_width = image->_width;
	_height = image->_height;
	// Allocate memory
	_img = (double *) calloc(_width*_height, sizeof(double));
	// Copy the contents
	memcpy(_img, image->_img,_width*_height*sizeof(double));
}

/**
 * @fn	CGrayscaleImage::CGrayscaleImage(unsigned char* img, size_t width, size_t height)
 *
 * @brief	Copy Constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]	img		The raw bytes of the image to be copied.
 * @param	width	   	The width.
 * @param	height	   	The height.
 */

CGrayscaleImage::CGrayscaleImage(unsigned char* img, size_t width, size_t height)
{
	// Allocate memory
	_img = (double *) calloc(width*height, sizeof(double));
	// Copy the contents (cannot do memcpy because img are unsigned char and I need doubles, so copying manually)
	for (unsigned int i = 0;i < width*height;i++)
	{
		_img[i] = img[i];
	}
	_width = width;
	_height = height;
}

/**
 * @fn	CGrayscaleImage::CGrayscaleImage(double *img, size_t width, size_t height)
 *
 * @brief	Copy Constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]	img		The input image.
 * @param	width	   	The width.
 * @param	height	   	The height.
 */

CGrayscaleImage::CGrayscaleImage(double *img, size_t width, size_t height)
{
	// Allocate memory
	_img = (double *) calloc(width*height, sizeof(double));
	// Copy contents
	memcpy(_img, img, width*height*sizeof(double));
	_width = width;
	_height = height;
}

/**
 * @fn	CGrayscaleImage::CGrayscaleImage(const char *filename)
 *
 * @brief	Constructor by filename. Reads an image from a file.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @exception	FILE_NOT_FOUND_EXCEPTION	Thrown when a FILE NOT FOUND error condition occurs.
 *
 * @param	filename	Filename of the file.
 */

CGrayscaleImage::CGrayscaleImage(const char *filename)
{
	// Try to open the image
	unsigned char *aux;
	if (NULL == (aux = read_png_u8_gray(filename, &_width, &_height)))
    {
		// If it doesn't exist, throw exception
		cerr<<"Image " << filename << " not read."<<endl;
		throw FILE_NOT_FOUND_EXCEPTION;
    }
	// If the image exists, allocate memory
	_img = (double *) calloc(_width*_height, sizeof(double));
	// Now copy the contents
	for (unsigned int i = 0;i < _width*_height;i++)
	{
		_img[i] = aux[i];
	}
	free(aux);
}

/**
 * @fn	CGrayscaleImage::~CGrayscaleImage(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CGrayscaleImage::~CGrayscaleImage(void)
{
	free(_img);
}

/**
 * @fn	double CGrayscaleImage::GetBilinearInterpolatedPixel(double x, double y)
 *
 * @brief	Gets bilinear interpolated pixel.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param	x	The x coordinate.
 * @param	y	The y coordinate.
 *
 * @return	The bilinear interpolated pixel.
 */

double CGrayscaleImage::GetBilinearInterpolatedPixel(double x, double y)
{
	// Calculate the four positions used to interpolate
	int xi = (int) x;
	int xi2 = xi + 1;
	int yi = (int) y;
	int yi2 = yi + 1;
	// Get the four positions
	double pxiyi = GETPIXEL(xi,yi);
	double pxi2yi = GETPIXEL(xi2,yi);
	double pxiyi2 = GETPIXEL(xi,yi2);
	double pxi2yi2 = GETPIXEL(xi2,yi2);
	double xx = x - xi;
	double yy = y - yi;

	// Perform the interpolation first horizontally and then interpolate both results vertically
	double val1 = pxiyi * (1-xx) + pxi2yi * (xx);
	double val2 = pxiyi2 * (1-xx) + pxi2yi2 * (xx);
	double res = val1 * (1-yy) + val2 * (yy);
	return res;
}

/**
 * @fn	double CGrayscaleImage::GetPixel(int x, int y)
 *
 * @brief	Gets a pixel.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param	x	The x coordinate.
 * @param	y	The y coordinate.
 *
 * @return	The pixel.
 */

double CGrayscaleImage::GetPixel(int x, int y)
{
	return _img[_width*y+x];
}

/**
 * @fn	void CGrayscaleImage::SetPixel(int x, int y, double value)
 *
 * @brief	Sets a pixel.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param	x	 	The x coordinate.
 * @param	y	 	The y coordinate.
 * @param	value	The value.
 */

void CGrayscaleImage::SetPixel(int x, int y, double value)
{
	_img[_width*y+x] = value;
}


/**
 * @fn	float* CGrayscaleImage::GetFloatVersion()
 *
 * @brief	Gets float version of image (Input image has values between 0 and 1).
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @return	Image in floating point ranging from 0 to 255
 */

float* CGrayscaleImage::GetFloatVersion()
{
	float *res = (float *) calloc(_width*_height, sizeof(float));
	for (unsigned int i=0;i<_width*_height;i++)
	{
		res[i] = (float)(_img[i] * 255);
		if (res[i] < 0) res[i] = 0;
		if (res[i] > 255) res[i] = 255;
	}
	return res;
}

/**
 * @fn	unsigned char* CGrayscaleImage::GetGrayscaleVersion()
 *
 * @brief	Get image in unsigned char values (Input image ranges from 0 to 255). This image is useful for saving or showing.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @return	Image in unsigned char values used to save or show.
 */

unsigned char* CGrayscaleImage::GetGrayscaleVersion()
{
	unsigned char *res = (unsigned char *) calloc(_width*_height, sizeof(unsigned char));
	for (unsigned int i=0;i<_width*_height;i++)
	{
		res[i] = (unsigned char) ROUND(_img[i]);
		if (_img[i] < 0) res[i] = 0;
		if (_img[i] > 255) res[i] = 255;
	}
	return res;
}

/**
 * @fn	void CGrayscaleImage::Convolution(CKernel &kernel, CGrayscaleImage *res)
 *
 * @brief	Performs image convolution. Not in-place, so the parameter $res$ can point to the same object (this). Ex: this.Convolution(kernel, this) is possible.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in]	kernel	The convolution kernel.
 * @param [out]	res   	The resulted convoluted image with the kernel.
 */

void CGrayscaleImage::Convolution(CKernel &kernel, CGrayscaleImage *res)
{
	// Allocate memory for storing the result
	CGrayscaleImage *out = new CGrayscaleImage(_width, _height);

	// find center position of kernel (half of kernel size)
	int kWidth = kernel.GetWidth();
	int kHeight = kernel.GetHeight();

	int kCenterX = kWidth / 2;
	int kCenterY = kHeight / 2;

	for(unsigned int i=0; i < _height; ++i)              // rows
	{
		for(unsigned int j=0; j < _width; ++j)          // columns
		{
			double acum = 0;
			for(int m=0; m < kHeight; ++m)     // kernel rows
			{
				int mm = kHeight - 1 - m;      // row index of flipped kernel
				for(int n=0; n < kWidth; ++n) // kernel columns
				{
					int nn = kWidth - 1 - n;  // column index of flipped kernel
					// index of input signal, used for checking boundary
					int ii = i + (m - kCenterY);
					int jj = j + (n - kCenterX);

					// ignore input samples which are out of bound
					if( ii >= 0 && ii < (int) _height && jj >= 0 && jj < (int) _width )
					{
						out->SetPixel(j,i,out->GetPixel(j,i) + GetPixel(jj, ii) * kernel.GetValue(nn,mm));
						acum += kernel.GetValue(nn,mm);
					}
				}
			}
			// Perform normalized averaging
			out->SetPixel(j,i, out->GetPixel(j,i) / acum);
		}
	}
	// Now copy the result in res.
	for (unsigned int x=0;x<_width;x++)
		for (unsigned int y=0;y<_height;y++)
		{
			double pix = out->GetPixel(x,y);
			res->SetPixel(x,y,pix);
		}
	delete out;
}