//
//  JavaScriptHelper.cpp
//

#include "Matching/DescriptorMatcher.h"
#include "Matching/DissimilarityScore.h"
#include "JavaScriptHelper.h"
#include <string>
#include <fstream>
#include <algorithm>
using namespace std;

/**
 * @fn	void printPixels(CMSER *mser, ofstream &myfile, int w1)
 *
 * @brief	Print pixels of MSER in file myfile, translated horizontally by w1 pixels
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param [in]	mser  		The mser.
 * @param [in,out]	myfile	The file.
 * @param	w1			  	Amount of pixles to translate horizontally. This could be used when the input MSER is shown in the second image on the right, so w1 should be the width of the first image.
 * 							If mser is on the first image, then it makes sense to set w1=0.
 */

void printPixels(CMSER *mser, ofstream &myfile, int w1)
{
	const vector<CPixel*>* tmp1=mser->GetPixels();
	for(unsigned int j=0;j<tmp1->size();j++)
	{
		int x=(int) ((*tmp1)[j]->first * pow(2.0, mser->GetScale()));
		int y=(int) ((*tmp1)[j]->second * pow(2.0, mser->GetScale()));
		if(j!=tmp1->size()-1)
			myfile<<x+w1<<","<<y<<",";
		else
			myfile<<x+w1<<","<<y;
	}
}

/**
 * @fn	void AddLineToLookupTable(int *lookUpTable, std::pair<CMSER *, CMSER *> &match, int w, int w1, int t)
 *
 * @brief	Adds each pixel of a match's line to a lookup table.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param [in,out]	lookUpTable	The look up table (image)
 * @param [in]	match	   		Specifies the match.
 * @param	w				   	The image width.
 * @param	w1				   	A value to translate the positions horizontally.
 * @param	t				   	The id to put in the lookup table.
 */

//void AddLineToLookupTable(int *lookUpTable, std::pair<CMSER *, CMSER *> &match, int w, int w1, int t)
//{
//    int x1=(int) (match.first->GetCentroid()->first * pow(2.0, match.first->GetScale()));
//    int y1=(int) (match.first->GetCentroid()->second * pow(2.0, match.first->GetScale()));
//    int x2=(int) (match.second->GetCentroid()->first * pow(2.0, match.second->GetScale()));
//    int y2=(int) (match.second->GetCentroid()->second * pow(2.0, match.second->GetScale()));
//    line(lookUpTable,w,x1-1,y1,x2+w1-1,y2,t);
//    line(lookUpTable,w,x1+1,y1,x2+w1+1,y2,t);
//    line(lookUpTable,w,x1,y1-1,x2+w1,y2-1,t);
//    line(lookUpTable,w,x1,y1+1,x2+w1,y2+1,t);
//    line(lookUpTable,w,x1,y1,x2+w1,y2,t);
//}

/**
 * @fn	void addMSER(CMSER &mser, std::map<int, int> &dictionary, vector<int> &addedMSERs, std::ofstream &myfile, bool isFirst, int w1)
 *
 * @brief	Adds a mser description to a file.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param [in]	mser	  		The mser.
 * @param [in]	dictionary		The dictionary.
 * @param [in,out]	addedMSERs	The already added MSERs
 * @param [in,out]	myfile	  	The file..
 * @param	isFirst			  	true if this object is first.
 * @param	w1				  	The amount to shift the MSER's pixels horizontally.
 */

void addMSER(CMSER &mser, std::map<int, int> &dictionary, vector<int> &addedMSERs, std::ofstream &myfile, bool isFirst, int w1)
{
	// If it does not exist, add it
	if (std::find(addedMSERs.begin(), addedMSERs.end(), mser.GetId()) == addedMSERs.end())
	{
		if(!isFirst)
			myfile<<","<<endl;
		myfile<<"[";
		printPixels(&mser, myfile, w1);
		addedMSERs.push_back(mser.GetId());
		dictionary[mser.GetId()] = (int) addedMSERs.size();
		myfile<<"]"<<endl;
	}
}

/**
 * @fn	void createJSFile(string nameFile,vector<std::pair<CMSER*,CMSER*> > shapeMatches,vector<std::pair<CMSER*,CMSER*> > textureMatches, vector<std::pair<CMSER*,CMSER*> > shapeMatchesIntermediate, vector<std::pair<CMSER*,CMSER*> > textureMatchesIntermediate, int w,int h,int w1)
 *
 * @brief	Creates javasript file to show in webpage with MSER and matches information.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param	nameFile						  	The name of the output file.
 * @param [in]	shapeMatches			  		The shape matches.
 * @param [in]	textureMatches			  		The texture matches.
 * @param	w								  	The width of the image
 * @param	h								  	The height of the image
 * @param	w1								  	A value to translate horizontally the position of the centroids.
 */

//void createJSFile(string nameFile,vector<std::pair<CMSER*,CMSER*> > shapeMatches,vector<std::pair<CMSER*,CMSER*> > textureMatches, vector<std::pair<CMSER*,CMSER*> > shapeMatchesIntermediate, vector<std::pair<CMSER*,CMSER*> > textureMatchesIntermediate, int w,int h,int w1)
//void createJSFile(string nameFile,vector<std::pair<CMSER*,CMSER*> > shapeMatches,vector<std::pair<CMSER*,CMSER*> > textureMatches, int w,int h,int w1)
void createJSFile(string nameFile,vector<std::pair<CMSER*,CMSER*> > shapeMatches,vector<std::pair<CMSER*,CMSER*> > textureMatches, int w1)
{
	// I will build a lookup table (an image) for both texture and shape descriptors in which each position has the id of the match going through it (i.e. the line between both MSERs centroids) or zero if no matching line passes through it.

	//int* shapeLookUpTable=new int[w*h];
    //memset(shapeLookUpTable,0,sizeof(int)*w*h);
    //int* textureLookUpTable=new int[w*h];
    //memset(textureLookUpTable,0,sizeof(int)*w*h);
	
    
    ofstream myfile;
    myfile.open (nameFile.c_str(),ios::trunc);

    // print the MSER
	// Add MSERs from all shape matches first
    myfile<<"var MSER1=["<<endl;
	vector<int> sourceMSERs = vector<int>();
	std::map<int, int> sourceIdToPos = std::map<int,int>();
    for(unsigned int i=0;i<shapeMatches.size();i++)
    {
		CMSER *mser = shapeMatches[i].first;
		addMSER(*mser, sourceIdToPos, sourceMSERs, myfile, i==0, 0);
    }
	// Now, for textures
    for(unsigned int i=0;i<textureMatches.size();i++)
    {
		CMSER *mser = textureMatches[i].first;
		addMSER(*mser, sourceIdToPos, sourceMSERs, myfile, false, 0);
    }
	// Uncomment this code if intermediate matches should be shown (higher transfer time)

	// Now, for intermediate shapes
	//for(unsigned int i=0;i<shapeMatchesIntermediate.size();i++)
	//{
	//	CMSER *mser = shapeMatchesIntermediate[i].first;
	//	addMSER(*mser, sourceIdToPos, sourceMSERs, myfile, false);
	//}
	// Now, for intermediate textures
	//for(unsigned int i=0;i<textureMatchesIntermediate.size();i++)
 //   {
	//	CMSER *mser = textureMatchesIntermediate[i].first;
	//	addMSER(*mser, sourceIdToPos, sourceMSERs, myfile, false);
 //   }


    myfile<<"];"<<endl;
	vector<int> targetMSERs = vector<int>();
	std::map<int, int> targetIdToPos = std::map<int,int>();

	myfile<<"var MSER2=["<<endl;
	for (unsigned int i=0;i<shapeMatches.size();i++)
	{
		CMSER *mser = shapeMatches[i].second;
		addMSER(*mser, targetIdToPos, targetMSERs, myfile, i==0, w1);
	}
	for (unsigned int i=0;i<textureMatches.size();i++)
	{
		CMSER *mser = textureMatches[i].second;
		addMSER(*mser, targetIdToPos, targetMSERs, myfile, false, w1);
	}

	// Uncomment this code if intermediate matches should be shown (higher transfer time)
	//for (unsigned int i=0;i<shapeMatchesIntermediate.size();i++)
	//{
	//	CMSER *mser = shapeMatchesIntermediate[i].second;
	//	addMSER(*mser, targetIdToPos, targetMSERs, myfile, i>0);
	//}
	//for (unsigned int i=0;i<textureMatchesIntermediate.size();i++)
	//{
	//	CMSER *mser = textureMatchesIntermediate[i].second;
	//	addMSER(*mser, targetIdToPos, targetMSERs, myfile, false);
	//}


    myfile<<"];"<<endl;

	// For Matches
	// Print of matches centroids
	myfile << "var shapeMatches=["<<endl;
    for (unsigned int i=0;i<shapeMatches.size();i++)
    {
        int x1=(int) (shapeMatches[i].first->GetCentroid()->first * pow(2.0, shapeMatches[i].first->GetScale()));
        int y1=(int) (shapeMatches[i].first->GetCentroid()->second * pow(2.0, shapeMatches[i].first->GetScale()));
        int x2=(int) (shapeMatches[i].second->GetCentroid()->first * pow(2.0, shapeMatches[i].second->GetScale()));
        int y2=(int) (shapeMatches[i].second->GetCentroid()->second * pow(2.0, shapeMatches[i].second->GetScale()));
		myfile << "[" << x1 << "," << y1 << "," << x2 + w1 << "," << y2 << "," << sourceIdToPos[shapeMatches[i].first->GetId()] << "," << targetIdToPos[shapeMatches[i].second->GetId()] << "]" << endl;
        if(i!=shapeMatches.size()-1) myfile << "," << endl;
	}
	myfile << "];"<<endl;

    
	// For Textures
	// Print of matches centroids
	myfile << "var textureMatches=["<<endl;
    for(unsigned int i=0;i<textureMatches.size();i++)
    {
        int x1=(int) (textureMatches[i].first->GetCentroid()->first * pow(2.0, textureMatches[i].first->GetScale()));
        int y1=(int) (textureMatches[i].first->GetCentroid()->second * pow(2.0, textureMatches[i].first->GetScale()));
        int x2=(int) (textureMatches[i].second->GetCentroid()->first * pow(2.0, textureMatches[i].second->GetScale()));
        int y2=(int) (textureMatches[i].second->GetCentroid()->second * pow(2.0, textureMatches[i].second->GetScale()));
		myfile << "[" << x1 << "," << y1 << "," << x2 + w1 << "," << y2 << "," << sourceIdToPos[textureMatches[i].first->GetId()] << "," << targetIdToPos[textureMatches[i].second->GetId()] << "]" << endl;
        if(i!=textureMatches.size()-1) myfile << "," << endl;
	}
	myfile << "];"<<endl;

	// Uncomment this code if it is necesary to print intermediate matches (unfiltered) 

	// For Matches Intermediate
	// Print of matches centroids
	//myfile << "var shapeMatchesIntermediate=["<<endl;
	//for (unsigned int i=0;i<shapeMatchesIntermediate.size();i++)
	//{
	//	int x1=(int) (shapeMatchesIntermediate[i].first->GetCentroid()->first * pow(2.0, shapeMatchesIntermediate[i].first->GetScale()));
	//	int y1=(int) (shapeMatchesIntermediate[i].first->GetCentroid()->second * pow(2.0, shapeMatchesIntermediate[i].first->GetScale()));
	//	int x2=(int) (shapeMatchesIntermediate[i].second->GetCentroid()->first * pow(2.0, shapeMatchesIntermediate[i].second->GetScale()));
	//	int y2=(int) (shapeMatchesIntermediate[i].second->GetCentroid()->second * pow(2.0, shapeMatchesIntermediate[i].second->GetScale()));
	//	myfile << "[" << x1 << "," << y1 << "," << x2 + w1 << "," << y2 << "," << sourceIdToPos[shapeMatchesIntermediate[i].first->GetId()] << "," << targetIdToPos[shapeMatchesIntermediate[i].second->GetId()] << "]" << endl;
	//	if(i!=shapeMatchesIntermediate.size()-1) myfile << "," << endl;
	//}
	//myfile << "];"<<endl;

    
	// For Textures Intermediate
	// Print of matches centroids
	//myfile << "var textureMatchesIntermediate=["<<endl;
	//for(unsigned int i=0;i<textureMatchesIntermediate.size();i++)
	//{
	//	int x1=(int) (textureMatchesIntermediate[i].first->GetCentroid()->first * pow(2.0, textureMatchesIntermediate[i].first->GetScale()));
	//	int y1=(int) (textureMatchesIntermediate[i].first->GetCentroid()->second * pow(2.0, textureMatchesIntermediate[i].first->GetScale()));
	//	int x2=(int) (textureMatchesIntermediate[i].second->GetCentroid()->first * pow(2.0, textureMatchesIntermediate[i].second->GetScale()));
	//	int y2=(int) (textureMatchesIntermediate[i].second->GetCentroid()->second * pow(2.0, textureMatchesIntermediate[i].second->GetScale()));
	//	myfile << "[" << x1 << "," << y1 << "," << x2 + w1 << "," << y2 << "," << sourceIdToPos[textureMatchesIntermediate[i].first->GetId()] << "," << targetIdToPos[textureMatchesIntermediate[i].second->GetId()] << "]" << endl;
	//	if(i!=textureMatchesIntermediate.size()-1) myfile << "," << endl;
	//}
	//myfile << "];"<<endl;

    myfile.close();
    
}

/**
 * @fn	void line(int* img, int w, int x1, int y1, int x2, int y2,int id)
 *
 * @brief	Draws a line on an image storing the value $id$ for this line.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param [in,out]	img	If non-null, the image.
 * @param	w		   	The width.
 * @param	x1		   	The first x value.
 * @param	y1		   	The first y value.
 * @param	x2		   	The second x value.
 * @param	y2		   	The second y value.
 * @param	id		   	The identifier.
 */

void line(int* img, int w, int x1, int y1, int x2, int y2,int id)
{
    
	int n, deltax, deltay, sgndeltax, sgndeltay, deltaxabs, deltayabs, x, y, drawx, drawy;
    
	deltax = x2 - x1;
	deltay = y2 - y1;
	deltaxabs = (int) fabs((double) deltax);
	deltayabs = (int) fabs((double) deltay);
	sgndeltax = deltax<0?-1:deltax>0;
	sgndeltay = deltay<0?-1:deltay>0;
	x = deltayabs >> 1;
	y = deltaxabs >> 1;
	drawx = x1;
	drawy = y1;
    
	img[drawx + drawy * w] = id;
    
	if(deltaxabs >= deltayabs){
        for(n = 0; n < deltaxabs; n++)
        {
            y += deltayabs;
            if(y >= deltaxabs)
            {
                y -= deltaxabs;
                drawy += sgndeltay;
            }
            drawx += sgndeltax;
            img[drawx+ drawy*w] = id;
        }
	}
	else
	{
		for(n = 0; n < deltayabs; n++)
		{
			x += deltaxabs;
			if(x >= deltayabs)
			{
				x -= deltayabs;
				drawx += sgndeltax; 
			} 
			drawy += sgndeltay; 
			img[drawx+ drawy*w] = id;
		} 
	} 
}