/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\Kernel.cpp
 *
 * @brief	Implements the kernel class. Responsible for representing a kernel and loading it with values (Gaussian for the moment)
 */

#include "Kernel.h"
#include <math.h>
#include "HelperMethods.h"
#include <stdlib.h>
#include <iostream>

using namespace std;

/**
 * @fn	CKernel::CKernel(int width, int height)
 *
 * @brief	Kernel Constructor.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param	width 	The kernel width.
 * @param	height	The kernel height.
 */

CKernel::CKernel(int width, int height) 
{
	_width = width;
	_height = height;
	// Build a 2D array to hold kernel values
	_data= new double*[height];
	for (int i=0;i<height;i++)
	{
		_data[i] = new double[width];
	}
}

/**
 * @fn	void CKernel::InitializeGaussian1DKernel(double sigma)
 *
 * @brief	Initializes a Gaussian 1-D kernel.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param	sigma	The standard deviation.
 */

void CKernel::InitializeGaussian1DKernel(double sigma)
{
	if (_width == 1 || _height == 1)
	{
		int sz = (_width == 1 ? _height : _width);
		int distToCenter = sz / 2;
		double center = (sz % 2 == 0?0.5:0);
		double sigma2 = sigma * sigma;
		for (int i=0;i<sz;i++)
		{
			double x = center - (double) distToCenter + (double) i;
			double value = exp(-1.0*((double)(x*x)/(2.0*sigma2))) / sqrt(2.0 * M_PI * sigma2);
			if (_width == 1)
				_data[i][0] = value;
			else
				_data[0][i] = value;
		}
	}
	else
	{
		cerr << "Kernel does not have 1D dimensions. Either its width or height must be 1." << endl;
	}
}

/**
 * @fn	void CKernel::InitializeGaussian2DKernel(double sigma)
 *
 * @brief	Initializes a Gaussian 2-D kernel.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param	sigma	Standard deviation.
 */

void CKernel::InitializeGaussian2DKernel(double sigma)
{
	int centerX = _width / 2;
	int centerY = _height / 2;
	double sigma2 = sigma * sigma;
	for (int x=-centerX;x<=centerX;x++)
	{
		for (int y=-centerY;y<=centerY;y++)
		{
			int xx = x + centerX;
			int yy = y + centerY;
			double value = exp(-1*((x*x+y*y)/(2*sigma2))) / (2 * M_PI * sigma2);
			_data[yy][xx] = value;
		}
	}
}

/**
 * @fn	double CKernel::GetValue(int x, int y)
 *
 * @brief	Gets a value.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param	x	The x coordinate.
 * @param	y	The y coordinate.
 *
 * @return	The kernel's value at a position.
 */
double CKernel::GetValue(int x, int y)
{
	return _data[y][x];
}

/**
 * @fn	CKernel::~CKernel(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	20/02/2014
 */

CKernel::~CKernel(void)
{
	for (int i=0;i<_height;i++)
	{
		delete[] _data[i];
	}
	delete[] _data;
}
