/**
 * @file
 * @author  Edouard Oyallon <edouard.oyallon@ens-cachan.fr>
 * @version 09/30/11
 *
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * http://www.gnu.org/copyleft/gpl.html
 *
 * @section DESCRIPTION
 *
 * This file contains some elements dedicated to build the MSER tree. No implementation of the build of descriptors has been done.
 */
#pragma once

/// Abs : usual absolute value
#define ABS(a) (((a)>0)?(a):-(a))
#ifndef MAX
#define MAX(x,y) (x) > (y) ? (x) : (y)
#endif
#ifndef MIN
#define MIN(x,y) (x) < (y) ? (x) : (y)
#endif

/// Delta is the delta of the Salembier algorithms. It refers to the difference between I(x+delta) and I(x-delta) so that we take a MSER

#include <complex>
#include <vector>
#include <iostream>
#include <string.h>
#include <math.h>
#include "../IO_PNG/io_png.h"
#include "../Params.h"
#include <set>

using namespace std;



#include "tree.h"


/**
 * Here is the class which is supposed to represent a MSER 
 */
class MSER
{
	public :
	// Positive tree or negative tree
	bool pos;
	// Sums with explicit names
	float sum_x,sum_y,sum_square_x, sum_square_y, sum_xy,area;
	/// Angle for the representation of ellipse
	float phi;
	/// First eigen value
	float eigen1,eigen2;
    
	/// The node of the MSER
	node* currentNode;
	/// The vector.
	float* vect;

    // Constructor
	MSER(node *theNode,int w)
	{	
		currentNode=theNode;
		sum_x=0;
		sum_y=0;
		sum_square_x=0;
		sum_square_y=0;
		sum_xy=0;
		area=(float) theNode->area;
		first_order(theNode,this,w);
		second_order(theNode,this,w,theNode->area);
		
        float a,b,c,dc;// Auxiliar variables
        
        a=sum_square_x;
		c=sum_square_y;
		b=sum_xy;
		eigen1=((a+c)+sqrt((a-c)*(a-c)+4*b*b))/2;
		eigen2=((a+c)-sqrt((a-c)*(a-c)+4*b*b))/2;
        dc = sqrt((a-c)*(a-c)+4*b*b);

		phi = atan2f( 0.5f*(c - a + dc) , b ) ;
				
		
	if(eigen1!=0 && eigen2!=0)
	 {
		
         // Here we compute the descriptors of a given MSER.
     }
	}

	
	/**
	 * Compute first order. It's recursive.
	 * @param n A node(lower or equal in level of the initial called)
	 * @param m The MSER related
	 * @param w the width of the image
	 */	
	void first_order(node *n, MSER* m,int w)
	{
		list_node* sons=n->sons;
		for(int i=0;i<n->list.size;i++)
		{
			m->sum_x+=(n->list.data)[i]%w;
			m->sum_y+=(n->list.data)[i]/w;
		}
	
		
		while(sons!=NULL)
		{
			first_order(sons->current_node,m,w);
			sons=sons->next_node;
		}
	}
	
	
	/**
	 * Compute first order. It's recursive.
	 * @param n A node(lower or equal in level of the initial called)
	 * @param m The MSER related
	 * @param w the width of the image
	 * @param area The area of the image
	 */	
	void second_order(node* n,MSER* m,int w,int area)
	{
		list_node* sons=n->sons;
		for(int i=0;i<n->list.size;i++)
		{m->sum_square_x+=((n->list.data)[i]%w-m->sum_x/area) * ((n->list.data)[i]%w-m->sum_x/area)/area;
			m->sum_square_y+=((n->list.data)[i]/w-m->sum_y/area) * ((n->list.data)[i]/w-m->sum_y/area)/area;
			m->sum_xy+=((n->list.data)[i]%w-m->sum_x/area) * ((n->list.data)[i]/w-m->sum_y/area)/area;
		}
		while(sons!=NULL)
		{
			second_order(sons->current_node,m,w,area);
			
			
			sons=sons->next_node;
		}
	}
	


};


/// Here is the list of MSER
typedef  std::vector<MSER*> listMSER;




/**
 * Incredible structure for a predefined list.
 * We have head x x x x x x x tail. When you look an element, the head goes to the tail. If you add an element, the tails go far from the head
 * @author Jerome Darbon
 */	
struct hqueue
{
	/// The first element.
	int head;
	/// The last element.
	int tail; /* First free place in queue, or -1 when the queue is full */
	/// The data
	int *data;
	hqueue() :  head(0), tail(0),data(NULL) {}
};



/// This macro returns the head, and increment the head.
#define hQueueFirst(hq) (hq.data[hq.head++])
/// This macro add an element to the tail.
#define hQueueAdd(hq, p) (hq.data[hq.tail++] = p)
/// This macro is used to add to the list of pixel a pixel at intensity i
#define addPixel(i,p) (pointerPixel[i].data[pointerPixel[i].size++] = p)

/// This macro says where is the list of pixel(it returns a pointer)
#define listPixel(i)(pointerPixel[i])

/// Put on the next part the composant.
#define next(i) pointerPixel[i].data+=pointerPixel[i].size;pointerPixel[i].size=0

/// Empty ?
#define hQueueEmpty(hq) (hq.head == hq.tail)
/// Number of colors.
#define H_MAX 256


/**
 * The MSER's description computes MSERs, fast. It's based on Salembier's article.
 */	


class MSER_description
{
private:
    
    int delta;
	int *img;// l'image
	bool* status;
	bool nodeAtLevel[H_MAX];
	int w,h,s;
	int histo[H_MAX];
	bool max_sel;
	
	
	int* listPixelNode;
	pixelSet pointerPixel[H_MAX];
	
	list_node* tree[H_MAX];
	
	
	node *root;
    hqueue m_queue[H_MAX];

	
    int** lookUP;
	// List of created nodes (used to free their memory later)
	vector<node*> nodes;
	
	
	
	public :

	listMSER* l;
	float* data;
    
    inline listMSER* getListMSER()
    {return l;}


	/**
	 * Recursive function to build the area of a domain
	 * @param n Node
	 */		
	void build_area(node* n)
	{
		if(n->sons==NULL)
		{
			n->area=n->list.size;
		}
		else
		{	
			int aux=n->list.size;
			list_node* sons=n->sons;
			while(sons != NULL)
			{	
				build_area(sons->current_node);
				aux+=(sons->current_node)->area;
				sons=sons->next_node;	
			}
			n->area=aux;
		}
		
	}
	
	
	/**
	 * The flood function. The heart of the program.
	 * @author Jerome Darbon
	 * @param n Level of flood
	 */	
	
	int flood(int n)
	{

		// Let's define the list of pixel.
	
		int p,q,i,lvl;

		hqueue& r = m_queue[n];
		
		while(!hQueueEmpty(r))
		{

			p=hQueueFirst(r);
			addPixel(n,p);
			status[p]=true;
			i=0;
			q=neighboor(p,i);
			while( q!=-1)
			{		 
				if(!status[q])
				{		
					lvl=img[q];
					hQueueAdd(m_queue[lvl],q);
					status[q]=true;
					nodeAtLevel[lvl]=true;
					
					while(lvl<img[p])
					{					
						lvl=flood(lvl);
						
					}
				}
				
				i++;
				q=neighboor(p,i);
			}
		}
		
		int m;
		m=n+1; 
		while(m<H_MAX && !nodeAtLevel[m])
			m++;
		
		

		if(m<H_MAX)
		{

			(tree[m]->current_node)->add_son(tree[n]->current_node);
			(tree[n]->current_node)->set_parent(tree[m]->current_node);
		}
		//root
		else {
			(tree[n]->current_node)->set_parent(NULL);
		}

		
		(tree[n]->current_node)->list=listPixel(n);
		next(n);

		tree[n]=tree[n]->add_node(n, nodes);
		
		nodeAtLevel[n]=false;
		
		return m;
	}
	
	/**
	 * This function return the neighboorhood point by point of the pixel p
	 * @param p Here is the pixel we focus on
	 * @param &i Here is the counter(dimension 2 -> 0<=i<4)
	 */
	int neighboor(int p,int &i)
	{
		
		if(i==0) 
		{if(p%w>0)
			return p-1;
		else
			return neighboor(p,++i);
		}
		if(i==1)
		{	if ( p>=w)
			return p-w;
		else 
			return neighboor(p,++i);
		}
		if(i==2)
		{if(p%w<w-1)
			return p+1;
		else 
			return neighboor(p,++i);
		}
		if(i==3)
		{if(p<=s-w-1)
			return p+w;
		else 
			return neighboor(p,++i);
		}
		return -1;
		
	}
	
	/**
	 * Initialize
	 * a It's the mode(upper/lower)
	 */	
	void init ()
	{

        status=new bool[s];
        
		// We prebuild the the tree and its root.
		for(int i=0;i<H_MAX;i++)
				tree[i]=new list_node(i, nodes);

		memset(histo,0,sizeof(int)*H_MAX);
        memset(nodeAtLevel,false,sizeof(bool)*H_MAX);
        memset(status,false,sizeof(bool)*s);


		int i_max=0;
		int max=img[0];
		


		for(int i=0;i<s;i++)
		{
			histo[img[i]]++;
			if(img[i]>max)
			{
				max=img[i];
				i_max=i;
			}	
		}	
		
		// Initialize lists
		listPixelNode=new int[s];
		int* m_queue_ptr = new int[s];
		int *offset = m_queue_ptr;
		int *offset2=listPixelNode;
		
		for (int i = 0; i < H_MAX; ++i)
		{
			int nb = histo[i];
			m_queue[i].data = offset;
			m_queue[i].head=0;
			m_queue[i].tail=0;
			pointerPixel[i].data=offset2;
			pointerPixel[i].size=0;
			offset += nb;
			offset2 += nb;
			
		}

		/// Initialization of the recursivity.
		nodeAtLevel[max] = true;
		hQueueAdd( m_queue[max],i_max);
		status[i_max] = true;

		
		
		////////////////////////////////// Important part
		flood(max);
        root=(tree[max]->next_node)->current_node;
        
        
		delete[] m_queue_ptr;
		
	}
	
	
	
	
	
	
	bool isParent(node *n1, node *n2)
	{
		if (n2->parent != NULL)
		{
			if (n2->parent == n1)
				return true;
			else
				return isParent(n1, n2->parent);
		}
		else
			return false;
	}

	bool areNested(node *n1, node *n2)
	{
		return isParent(n1, n2) || isParent(n2,n1);
	}
	void remove_MinDiversity()
	{
		for (int i=0;i<(int)l->size();i++)
		{
			node *n = l->at(i)->currentNode;
			for (int j=i;j<(int)l->size();j++)
			{
				if (i != j)
				{
					node *n2 = l->at(j)->currentNode;
					// Check if they are nested
					if (areNested(n,n2))
					{
						float minArea = static_cast<float>(MIN(n->area, n2->area));
						float maxArea = static_cast<float>(MAX(n->area, n2->area));
						float relAreaVariation = minArea / maxArea;
						if (relAreaVariation > CParams::Instance()->minDiversity)
						{
							// Keep the more stable one (the one with lower ratio)
							if (n2->ratio < n->ratio)
							{
								delete l->at(i);
								l->erase(l->begin() + i);
								n = l->at(i)->currentNode;
								j = -1;
							}
							else
							{
								delete l->at(j);
								l->erase(l->begin() + j);
								j--;
							}
						}
					}
				}
			}
		}
	}

	/**
	 * This function return the MSER, it takes in consideration if we have an Upper or a Downer tree
	 * @param n the tree. Initialize with the root
	 * @param pos True : up ; Down : false
	 */	
	void build_MSER(node* n)
	{


		bool isMser1=((n->parent != NULL) && ((n->parent)->ratio)>n->ratio)?true:false;
		
		//if (isMser1)
		//{
		//	float relAreaVariation = (float) n->area / (float) (n->parent)->area;
		//	if (relAreaVariation > 0.8)
		//		isMser1 = false;
		//}
		
		list_node* sons=n->sons;
		bool isMser2=(sons!=NULL);
		while(sons!=NULL)
		{
			build_MSER(sons->current_node);
				
			if((sons->current_node)->ratio<=n->ratio)
			{
				isMser2=false;
			}
			//else
			//{
			//	float relAreaVariation = (float) (sons->current_node)->area / (float) n->area;
			//	if (relAreaVariation > 0.8 && isMser1)
			//	{
			//		isMser2 = false;
			//	}
			//		
			//}
			sons=sons->next_node;

		}
		if(isMser2 && isMser1)
		{
			l->push_back(new MSER(n,w));
			if((l->back())->eigen1==0 || (l->back())->eigen2==0)
			{
				delete l->at(l->size()-1);
				l->pop_back();
			}
		}
		
		
	}
	
	
	/**
	 * This builds recursively the ratio \f$ #(I(n)+delta)-#(I(n)-delta)/2*delta*area\f$
	 * @param n The node..
	 */
	void build_ratio(node* n)
	{
		
		
		if(n->sons==NULL)
		{
			n->ratio=0;

				
		}
		else
		{	
			list_node* sons=n->sons;
			if(n->parent==NULL)
			{			n->ratio=0;
			}
			else
			{
			
				n->ratio=(float)(-iMinusDelta(n,n->intensity)+iPlusDelta(n,n->intensity))/(2*delta*(float)n->area);
				
			}
			while(sons != NULL)
			{	
			build_ratio(sons->current_node);
				sons=sons->next_node;	
				
			}
		}
	}
	
	
	
	
	/**
	 * Compute the area of the region so that I-DELTA is the intensity line ; compute the union of the area of the sons of the original call to this function
	 * @param i the intensity
	 * @param n the node
	 * @param mode_sup If it's an upper or downer tree
	 */
	int iMinusDelta(node* n,int i)
	{
		int area=0;
		if(n->intensity<=i-delta)
			return n->area;
		else 
		{	
			list_node* sons=n->sons;
			while(sons!=NULL)
			{
				area+=iMinusDelta(sons->current_node,i);
				sons=sons->next_node;
			}
			return area;
		}

		
	}
	
	/// Compute the area of the region in the level line of I+DELTA intensity.
	int iPlusDelta(node* n,int i)
	{
		
		// If there is no parent,
		if(n->parent==NULL)
			return n->area;
		if((n->parent)->intensity>i+delta)
		{
			return (n->parent)->area;
		}
		else 
			return iPlusDelta(n->parent,i);
		
	}
	
	

	~MSER_description()
	{
		delete[] status;
		std::set<node*> s = std::set<node*>();

		for(int i=0;i<H_MAX;i++)
		{
			delete tree[i];
			tree[i] = NULL;
		}
		for (unsigned int i=0;i<nodes.size();i++)
		{
			delete nodes[i];
		}
		// Delete all MSERs
		for (unsigned int i=0;i<l->size();i++)
		{
			delete l->at(i);
		}
		delete l;
		delete[] listPixelNode;
		
	}
	
	/**
	 * Constructor.
	 * @param *image the image...
	 * @param nx Width
	 * @param ny Height
	 */
	MSER_description(int *image,int nx, int ny,int delta_)
	{
        delta=delta_;
		w=nx;
		h=ny;
		s=nx*ny;
		img=image;
		
		
		//cout<<"***************************************************************"<<endl;
		init();

//		cout<<"Tree built."<<endl;
		
		l=new listMSER();
		
		// We define the tree, positive and negative.

		build_area(root);

//		cout<<"Areas of each node computed."<<endl;
		
		build_ratio(root);
//		cout<<"Ratios computed."<<endl;
		
		
		
		
		build_MSER(root);

//		cout<<"MSERs computed."<<endl;

		remove_MinDiversity();
//		cout << "Min Diversity check done" << endl;

//		cout<<"***************************************************************"<<endl<<endl;
	}
	
	
	
	
};	
