/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\MSERWrapper.cpp
 *
 * @brief	Implements the MSER wrapper class, to wrap Oyallon's MSER functionalities.
 */

#include "MSERWrapper.h"
#include "Params.h"
#include "MSER/mser.h"
#include <math.h>

/**
 * @fn	CMSERWrapper::CMSERWrapper(CGrayscaleImage &img)
 *
 * @brief	Constructor.
 *
 * @author	Tato
 * @date	21/02/2014
 *
 * @param [in,out]	img	The input image to calculate the MSERs
 */

CMSERWrapper::CMSERWrapper(CGrayscaleImage &img, int scale) : _scale(scale), _img(img)
{
	// First, convert the input image into Oyallon's MSER input image type
	int* img1=new int[img.Width()*img.Height()];
	for(int i=0;i<img.Width()*img.Height();i++)
	{
		img1[i]=(int) img.GetValue(i);
	}
	
	// Now instantiate his class to generate the MSERs.
	
	//_desc = new MSER_creation(img1,img.Width(),img.Height(), CParams::Instance()->delta);
	_desc = new MSER_description(img1,img.Width(),img.Height(), CParams::Instance()->delta);

	delete[] img1;

	// No visualization needed
	//_desc->visualisation_MSER("c:\\output1.png");
}

/**
 * @fn	CMSERList * CMSERWrapper::GenerateMSERList(listMSER *l, int scale)
 *
 * @brief	Generates the MSER list, taking into account the mininMSERArea parameter.
 *
 * @author	Tato
 * @date	21/02/2014
 *
 * @param [in]	l		The listMSER structure provided from Oyallon's classes to process.
 * @param	scale	 	The scale in which the MSERs were computed. Since the class only obtains a filename, 
 *
 * @return	null if it fails, else the mser list.
 */

CMSERList * CMSERWrapper::GenerateMSERList(listMSER *l)
{
	CMSERList *list = new CMSERList(l->size());
	int qty = 0;
	for (unsigned int i=0;i<l->size();i++)
	{
		MSER *m = l->at(i);
		// If the scaled area of the detected MSER is bigger than the threshold, add it to the list, otherwise ignore it.
		if (m->area * pow(2.0,(_scale*2.0)) > CParams::Instance()->minimumMSERArea)
		{
			CMSER *mser = new CMSER(m, _img.Width(), _img.Height(),  &_img, _scale, i);
			list->at(qty++) = mser;
		}
	}
	// Need to resize the list because there may have been some MSERs that did not have enough area.
	list->resize(qty);
	return list;
}

/**
 * @fn	CMSERWrapper::~CMSERWrapper(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	21/02/2014
 */

CMSERWrapper::~CMSERWrapper(void)
{
	delete _desc;
}


/**
 * @fn	CMSERList *CMSERWrapper::GetList(int scale)
 *
 * @brief	Gets a list. When asking for the list of MSER, remember to delete it and ALSO ITS MSERs!
 *
 * @author	Tato
 * @date	21/02/2014
 *
 * @param	scale	The scale in which the MSERs were computed.
 *
 * @return	The list if MSERs from the input image
 */

CMSERList *CMSERWrapper::GetList()
{
	return GenerateMSERList(_desc->getListMSER());
}
