/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\MatchesImageGenerator.cpp
 *
 * @brief	Implements the matches image generator class. Responsible for generating an image with two images next to each other and a line between the MSERs centroids connecting all matches.
 */

#include "MatchesImageGenerator.h"

/**
 * @fn	CMatchesImageGenerator::CMatchesImageGenerator(CGrayscaleImage &im1, CGrayscaleImage &im2)
 *
 * @brief	Constructor.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param [in,out]	im1	The first image.
 * @param [in,out]	im2	The second image.
 */

CMatchesImageGenerator::CMatchesImageGenerator(CGrayscaleImage &im1, CGrayscaleImage &im2) : _im1(im1), _im2(im2)
{
}

/**
 * @fn	void CMatchesImageGenerator::GenerateImage(const char *filename, vector<pair<CMSER*,CMSER*> > &matches)
 *
 * @brief	Generates an image with the two input images one next to each other (horizontally) and draw a line for each match.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param	filename	   	Name of the file.
 * @param [in]	matches		The computed matches between MSERs of both images.
 */

void CMatchesImageGenerator::GenerateImage(const char *filename, vector<pair<CMSER*,CMSER*> > &matches)
{
	int mx = _im1.Width();
	int nx = _im2.Width();
	int my = _im1.Height();
	int ny = _im2.Height();
	int y=(my>ny)?my:ny;
	int x=mx+nx;
	//for (int i=0;i<matches.size();i++)
	//{
	float* output=new float[x*y];
	for(int i=0;i<x;i++)
		for(int j =0;j<y;j++)
		{
			output[i+j*x] = 0;
		}
	
	
	// Paste the first image.
	for(int j=0;j<my;j++)
	{
		for(int i=0;i<mx;i++)
		{
			output[i+j*x]= (float) _im1.GetPixel(i,j);
		}
	}
	// Paste the second image.
	for(int j=0;j<ny;j++)
	{
		for(int i=mx;i<x;i++)
		{
			output[i+j*x]= (float) _im2.GetPixel(i-mx, j);
		}
	}

	// Now draw a line between both matches
	for(unsigned int i=0;i<matches.size();i+=1)
	{
		int x1= (int) (matches[i].first->GetCentroid()->first * pow(2.0, matches[i].first->GetScale()));
		int y1= (int) (matches[i].first->GetCentroid()->second * pow(2.0, matches[i].first->GetScale()));
		int x2=(int) (matches[i].second->GetCentroid()->first * pow(2.0, matches[i].second->GetScale()));
		int y2=(int) (matches[i].second->GetCentroid()->second * pow(2.0, matches[i].second->GetScale()));
		x1=(x1<mx)?x1:mx;
		x2=(x2<nx)?x2:nx;
		y1=(y1<my)?y1:my;
		y2=(y2<ny)?y2:ny;
		lign(output,x,x1,y1,x2+mx,y2);

	}
	// Write the filename
	write_png_f32(filename, output,x,y,1);
	delete[] output;
	//}
}

/**
 * @fn	CMatchesImageGenerator::~CMatchesImageGenerator(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	20/02/2014
 */

CMatchesImageGenerator::~CMatchesImageGenerator(void)
{
}

/**
 * @fn	void CMatchesImageGenerator::lign(float *img, int w, int x1, int y1, int x2, int y2)
 *
 * @brief	Draws a line on an image.
 *
 * @author	Tato
 * @date	20/02/2014
 *
 * @param [in,out]	img	The output image.
 * @param	w		   	The image's width.
 * @param	x1		   	The first x value.
 * @param	y1		   	The first y value.
 * @param	x2		   	The second x value.
 * @param	y2		   	The second y value.
 */

void CMatchesImageGenerator::lign(float *img, int w, int x1, int y1, int x2, int y2)
{

	int n, deltax, deltay, sgndeltax, sgndeltay, deltaxabs, deltayabs, x, y, drawx, drawy; 

	deltax = x2 - x1; 
	deltay = y2 - y1; 
	deltaxabs = (int) fabs((double) deltax); 
	deltayabs = (int) fabs((double) deltay); 
	sgndeltax = deltax<0?-1:deltax>0; 
	sgndeltay = deltay<0?-1:deltay>0; 
	x = deltayabs >> 1; 
	y = deltaxabs >> 1; 
	drawx = x1; 
	drawy = y1; 

	img[drawx + drawy * w] = 255; 

	if(deltaxabs >= deltayabs){ 
	for(n = 0; n < deltaxabs; n++)
	{ 
		y += deltayabs; 
		if(y >= deltaxabs)
		{ 
			y -= deltaxabs; 
			drawy += sgndeltay; 
		} 
		drawx += sgndeltax; 
		img[drawx+ drawy*w] = 255; 
	} 
	}
	else
	{ 
		for(n = 0; n < deltayabs; n++)
		{
			x += deltaxabs; 
			if(x >= deltayabs)
			{ 
				x -= deltayabs; 
				drawx += sgndeltax; 
			} 
			drawy += sgndeltay; 
			img[drawx+ drawy*w] = 255; 
		} 
	} 
}
