#include "DissimilarityScore.h"
#include <limits>
#include <iostream>
#include <algorithm>
#include <vector>

//#define AREADIF 300

CDissimilarityScore::CDissimilarityScore(void)
{
}


CDissimilarityScore::~CDissimilarityScore(void)
{
}

/**
 * @fn	double CDissimilarityScore::CompareDescriptors(double d1[128], double d2[128])
 *
 * @brief	Compare descriptors.
 *
 * @author	Tato
 * @date	23/02/2014
 *
 * @param	d1	The first double.
 * @param	d2	The second double.
 *
 * @return	A double.
 */

double CDissimilarityScore::CompareDescriptors(double d1[128], double d2[128])
{
	double acum = 0;
	for (int i=0;i<128;i++)
	{

		if (d1[i] + d2[i] != 0)
			acum += ((d1[i] - d2[i]) * (d1[i] - d2[i])) / (d1[i] + d2[i]);
	}
	return acum / 2.0;
}

/**
 * @fn	double CDissimilarityScore::GetScore(double descriptor[128], vector<std::pair<int,double*> > &db, int &bestMatch, const CMSERList *dbMSER)
 *
 * @brief	Gets a score.
 *
 * @author	Tato
 * @date	23/02/2014
 *
 * @param	descriptor		 	The descriptor.
 * @param [in,out]	db		 	[in,out] If non-null, the database.
 * @param [in,out]	bestMatch	A match specifying the best.
 * @param	dbMSER			 	The database mser.
 *
 * @return	The score.
 */

double CDissimilarityScore::GetScore(double descriptor[128], vector<std::pair<int,double*> > &db, int &bestMatch, const CMSERList *dbMSER)
{
	double d1 = numeric_limits<double>::max( );
	double d2 = numeric_limits<double>::max( );
	bestMatch = -1;
	int secondBest = -1;

	for (unsigned int i=0;i<db.size();i++)
	{
		double dist = CompareDescriptors(descriptor, db[i].second);
		if (dist < d1)
		{
			d2 = d1;
			d1 = dist;
			secondBest = bestMatch;
			bestMatch = db[i].first;
		}
		else if (dist < d2)// && bestMatch != db[i].first)
		{
			d2 = dist;
			secondBest = db[i].first;
		}
	}
	if (secondBest != -1 && bestMatch != -1)
	{
		CMSER *m1 = dbMSER->at(bestMatch);
		CMSER *m2 = dbMSER->at(secondBest);
		// If the centroids of both best and second best matches are the same and have similar areas, search for a second best match that either has different centroid either has different area
		// Store the MSERs that have already been tried
		vector<int> alreadyTriedMSERs = vector<int>();
		while (m1->HasSameCentroid(*m2, true) && m1->HasSimilarArea(*m2))
		{
			// Add to the already tried vector 
			alreadyTriedMSERs.push_back(secondBest);
			// Now search for the second best again, excluding the already tried ones
			double d2 = numeric_limits<double>::max( );
			for (unsigned int i=0;i<db.size();i++)
			{
				int idToFind = db[i].first;
				typedef vector<int>::iterator IntIterator;
				IntIterator result = std::find(alreadyTriedMSERs.begin(), alreadyTriedMSERs.end(), idToFind);
				bool notAlreadyTried = (result == alreadyTriedMSERs.end());
				if (notAlreadyTried)
				{
					double dist = CompareDescriptors(descriptor, db[i].second);
					if (dist < d2)
					{
						d2 = dist;
						secondBest = db[i].first;
					}
				}
			}
			if (d2 != numeric_limits<double>::max())
				m2 = dbMSER->at(secondBest);
			else
				break;
		}
	}
	// Return the score if a second best was found, otherwise the score is 0 (it there is not second best, so the first best should be always assumed as a match)
	if (d2 != 0 && d2 != numeric_limits<double>::max( ))
		return d1/d2;
	else if (d2 == 0)
		return numeric_limits<double>::max( );
	else if (d1 != numeric_limits<double>::max( ))
		return 0;
	else // If first and second best matches not found, return $\infty$
		return numeric_limits<double>::max( );
}
