/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\MserPlusSiftMain.cpp
 *
 * @brief	Implements the mser plus sift main class.
 */

#include "MserPlusSiftMain.h"
#include <assert.h>
#include <iostream>
#include "IO_PNG/io_png.h"
#include "MultiscaleMSER.h"
#include "ShapeDescriptor.h"
#include "TextureDescriptor.h"
#ifdef _OPENMP
#include <omp.h>
#endif
#include "MatchesImageGenerator.h"
#include "Params.h"
#include "ShapeImageGenerator.h"
#include "TextureImageGenerator.h"
#include "Matching/DescriptorMatcher.h"
#include <cstdlib>
#include "JavaScriptHelper.h"
#include "RansacWrapper.h"
#include <stdio.h>

using namespace std;

int CMserPlusSiftMain::GetMSERSiftFeatures(const int descriptorType, char *image, int delta, const CMSERList **msers, vector<vector<double*>* > **mserDescriptors, vector<CGrayscaleImage * > **mserRotations, vector<std::pair<double, double> > **mserTranslations, double *scale)
{
	CGrayscaleImage *img1 = NULL;
	try
	{
		img1 = new CGrayscaleImage(image);
	}
	catch (int n)
	{
		if (n == FILE_NOT_FOUND_EXCEPTION)
			cout << "File " << image << " not found" << endl;
		return EXIT_FAILURE;
	}
	CParams::Instance()->delta = delta;
	// Extracting the MSER
	cout << "Extracting MSERs from image." << endl;
	int scales = CParams::Instance()->qtyScales;
	CMultiscaleMSER msMSER = CMultiscaleMSER(img1, scales);
	const CMSERList *listMSER = msMSER.GetMSERList();
	*msers = listMSER;
	//tempDescs = vector<std::pair<int, vector<double*> > >(listMSER->size());
	*mserDescriptors = new vector<vector<double*>* >(listMSER->size());
	*mserRotations = new vector<CGrayscaleImage * >(listMSER->size());
	*mserTranslations = new vector< std::pair<double, double> > (listMSER->size());
	// First extract all descriptors
	int qtyDescs = 0;
	for (int i = 0; i<(int)listMSER->size(); i++)
	{
		CDescriptor *descriptor;
		if (descriptorType == CMserPlusSiftMain::MSERShape)
		{
			CParams::Instance()->SetDefaultShapeParams();
			descriptor = new CShapeDescriptor();
		}
		else if (descriptorType == CMserPlusSiftMain::MSERTexture)
		{
			CParams::Instance()->SetDefaultTextureParams();
			descriptor = new CTextureDescriptor();
		}
		else
		{
			cout << "Wrong parameter for descriptor type" << endl;
			return EXIT_FAILURE;
		}
		// Perform the resampling and obtain the SIFT descriptor
		CMSER *m = listMSER->at(i);
		descriptor->Init(*m, *img1);
		//vector<double*> descriptors = vector<double*>();
		// Add the ID of the MSER and all SIFT descriptors for it to the temporal vector
		//tempDescs[i] = std::pair<int, vector<double*> >(i, descriptors);
		//qtyDescs += (int)descriptors.size();

		// Descriptors
		(*mserDescriptors)->at(i) = new vector<double *>();
		descriptor->CalculateSIFTDescriptor(*((*mserDescriptors)->at(i)));
		qtyDescs += (int) ((*mserDescriptors)->at(i))->size();
		// Save the transformation
		// Rotation
		CGrayscaleImage *rotMatrix = descriptor->GetRotationMatrix();
		(*mserRotations)->at(i) = new CGrayscaleImage(rotMatrix);
		// Scale
		*scale = CParams::Instance()->s;
		// Translation
		(*mserTranslations)->at(i) = std::pair<double, double>(descriptor->GetTranslation());
		delete descriptor;
	}

	if (img1 != NULL)
		delete img1;
	return qtyDescs;
}

/**
 * @fn	int CMserPlusSiftMain::DoMSERPlusSift(char *file1, char *file2, int delta, double matchThreshold)
 *
 * @brief	Compute all filtered matches using both a texture descriptor and a shape descriptor based on MSERs.
 *
 * @author	Tato
 * @date	18/02/2014
 *
 * @param [in,out]	file1 	The first file.
 * @param [in,out]	file2 	The second file.
 * @param	delta		  	The delta (MSER parameter).
 * @param	matchThreshold	The match threshold (match acceptance rate).
 *
 * @return	EXIT_SUCCESS or EXIT_FAILURE.
 */

int CMserPlusSiftMain::DoMSERPlusSift(char *file1, char *file2, int delta, double matchThreshold)
{
	CGrayscaleImage *img1 = NULL;
	CGrayscaleImage *img2 = NULL;
	try
	{
		img1 = new CGrayscaleImage(file1);
	}
	catch(int n)
	{
		if (n == FILE_NOT_FOUND_EXCEPTION)
			cout << "File " << file1 << " not found" << endl;
		return EXIT_FAILURE;
	}
    try
	{
		img2 = new CGrayscaleImage(file2);
	}
	catch(int n)
	{
		if (n == FILE_NOT_FOUND_EXCEPTION)
			cout << "FIle " << file2 << " not found" << endl;
		return EXIT_FAILURE;
	}
	CParams::Instance()->delta = delta;
	CParams::Instance()->matchThreshold = matchThreshold;

#ifdef _OPENMP
    const double mserStart = omp_get_wtime();
#endif
    // Extracting the MSER
	int scales = CParams::Instance()->qtyScales;
	cout << "Extracting MSERs from first image." << endl;
	CMultiscaleMSER msMSER = CMultiscaleMSER(img1, scales);
	const CMSERList *listMSER1 = msMSER.GetMSERList();
	cout << "Extracting MSERs from second image." << endl;
	CMultiscaleMSER msMSER2 = CMultiscaleMSER(img2, scales);
	const CMSERList *listMSER2 = msMSER2.GetMSERList();
#ifdef _OPENMP
    const double mserEnd = omp_get_wtime();
#endif

	// Now, with each MSER, I will resample it to create its shape descriptor
	// Do it in parallel using OPENMP
	int i;
    
    // Shape and texture descriptors
	vector<std::pair<int, vector<double*> > > descTempShapes = vector<std::pair<int, vector<double*> > >(listMSER1->size());
	vector<std::pair<int, vector<double*> > > descTempTextures = vector<std::pair<int, vector<double*> > >(listMSER1->size());
	int qtyShapes = 0, qtyTextures = 0;
	CParams::Instance()->SetDefaultShapeParams();
	
#ifdef _OPENMP
    const double shape1Start = omp_get_wtime();
#endif

	CShapeImageGenerator sig = CShapeImageGenerator((int) listMSER1->size(), true);

	// Now for all MSERs on image 1, compute the shape descriptors
	#pragma omp parallel for ordered schedule(dynamic) reduction(+:qtyShapes)
	for (i=0;i<(int) listMSER1->size();i++)
	{
		CMSER *m = listMSER1->at(i);
		CShapeDescriptor shapeDescriptor = CShapeDescriptor();
		shapeDescriptor.Init(*m, *img1);
		float *shapeImg = shapeDescriptor.GetAffineInvariantImage()->GetFloatVersion();
		sig.addShape(shapeImg, i);
		vector<double*> descs = vector<double*>();
		shapeDescriptor.CalculateSIFTDescriptor(descs);
		descTempShapes[i] = std::pair<int, vector<double*> >(i,descs);
		qtyShapes += (int) descs.size();
	}
	cout << "Amount of shape descriptors image 1: " << qtyShapes << " ";
#ifdef _OPENMP	
    const double shape1End = omp_get_wtime();
	cout<<"("<<(shape1End)-(shape1Start)<<"s)"<<endl;
#endif


#ifdef _OPENMP
    const double texture1Start = omp_get_wtime();
#endif
	// Now for all MSERs on image 1, compute the texture descriptors
    CTextureImageGenerator textures = CTextureImageGenerator((int) listMSER1->size(), true);
	CParams::Instance()->SetDefaultTextureParams();
	#pragma omp parallel for ordered schedule(dynamic) reduction(+:qtyTextures)
	for (i=0;i<(int) listMSER1->size();i++)
	{
		CMSER *m = listMSER1->at(i);
		CTextureDescriptor textureDescriptor = CTextureDescriptor();
		textureDescriptor.Init(*m, *img1);
		textures.addTexture(textureDescriptor.GetAffineInvariantImage()->GetGrayscaleVersion(), i);
		vector<double*> descs2 = vector<double*>();
		textureDescriptor.CalculateSIFTDescriptor(descs2);
		descTempTextures[i] = std::pair<int, vector<double*> >(i,descs2);
		qtyTextures += (int) descs2.size();
	}
	cout << "Amount of texture descriptors image 1: " << qtyTextures << " ";
#ifdef _OPENMP
    const double texture1End = omp_get_wtime();
	cout<<"("<<(texture1End)-(texture1Start)<<"s)"<<endl;
#endif

	cout << "Generating images for both shapes and textures for image 1 ";
	// Now generate the image where all shapes/textures are shown for image 1

#ifdef _OPENMP
    const double image1ShapesAndTexturesStart = omp_get_wtime();
#endif
	sig.GenerateImage("shapes1.png");
	textures.GenerateImage("textures1.png");
#ifdef _OPENMP
    const double image1ShapesAndTexturesEnd = omp_get_wtime();
	cout<<"("<<(image1ShapesAndTexturesEnd)-(image1ShapesAndTexturesStart)<<"s)"<<endl;
#endif

	// Since there could be more than a descriptor per keypoint (more than one principal orientation), 
	// put all shape/texture descriptors together in the descI1Shapes/descI1Textures vector, where each element of the vector is a pair
	// where the first index is the MSER's id and the second the descriptor itself (2^8 values)
	vector<std::pair<int, double*> > descI1Shapes = vector<std::pair<int, double*> >(qtyShapes);
	vector<std::pair<int, double*> > descI1Textures = vector<std::pair<int, double*> >(qtyTextures);
	qtyShapes = 0;
	qtyTextures = 0;
	for (unsigned int i=0;i<descTempShapes.size();i++)
	{
		for (unsigned int j=0;j<descTempShapes[i].second.size();j++)
		{
			descI1Shapes[qtyShapes].first = descTempShapes[i].first;
			descI1Shapes[qtyShapes++].second = descTempShapes[i].second.at(j);
		}
	}
	for (unsigned int i=0;i<descTempTextures.size();i++)
	{
		for (unsigned int j=0;j<descTempTextures[i].second.size();j++)
		{
			descI1Textures[qtyTextures].first = descTempTextures[i].first;
			descI1Textures[qtyTextures++].second = descTempTextures[i].second.at(j);
		}
	}
	// Do the same for the second image
	//cout << "Computing shape descriptors for second image" << endl;
	qtyShapes = 0;
	qtyTextures = 0;
	descTempShapes.clear();
	descTempTextures.clear();
	descTempShapes.resize(listMSER2->size());
	descTempTextures.resize(listMSER2->size());

	CParams::Instance()->SetDefaultShapeParams();
#ifdef _OPENMP
    const double shape2Start = omp_get_wtime();
#endif
	CShapeImageGenerator sig2 = CShapeImageGenerator((int) listMSER2->size(), true);
	// Now for all MSERs on image 2, compute the shape descriptors
	#pragma omp parallel for ordered schedule(dynamic) reduction(+:qtyShapes)
	for (i=0;i<(int) listMSER2->size();i++)
	{
		CMSER *m = listMSER2->at(i);
		CShapeDescriptor shapeDescriptor = CShapeDescriptor();
		shapeDescriptor.Init(*m, *img2);
		sig2.addShape(shapeDescriptor.GetAffineInvariantImage()->GetFloatVersion(), i);
		vector<double*> descs = vector<double*>();
		shapeDescriptor.CalculateSIFTDescriptor(descs);
		descTempShapes[i] = std::pair<int, vector<double*> >(i,descs);
		qtyShapes+= (int) descs.size();
	}
	cout << "Amount of shape descriptors image 2: " << qtyShapes << " ";
#ifdef _OPENMP
    const double shape2End = omp_get_wtime();
	cout<<"("<<(shape2End)-(shape2Start)<<"s)"<<endl;
#endif

	// Now for all MSERs on image 2, compute the texture descriptors

#ifdef _OPENMP
	const double texture2Start = omp_get_wtime();
#endif
    CTextureImageGenerator textures2 = CTextureImageGenerator((int) listMSER2->size(), true);

	CParams::Instance()->SetDefaultTextureParams();
	#pragma omp parallel for ordered schedule(dynamic) reduction(+:qtyTextures)
	for (i=0;i<(int) listMSER2->size();i++)
	{
		CMSER *m = listMSER2->at(i);
		CTextureDescriptor textureDescriptor = CTextureDescriptor();
		textureDescriptor.Init(*m, *img2);
		textures2.addTexture(textureDescriptor.GetAffineInvariantImage()->GetGrayscaleVersion(), i);

		vector<double*> descs2 = vector<double*>();
		textureDescriptor.CalculateSIFTDescriptor(descs2);
		descTempTextures[i] = std::pair<int, vector<double*> >(i,descs2);
		qtyTextures += (int) descs2.size();

	}
    
	cout << "Amount of texture descriptors image 2: " << qtyTextures << " ";
#ifdef _OPENMP
	const double texture2End = omp_get_wtime();
	cout<<"("<<(texture2End)-(texture2Start)<<"s)"<<endl;
#endif

	cout << "Generating images for both shapes and textures for image 2 ";

#ifdef _OPENMP
    const double image2ShapesAndTexturesStart = omp_get_wtime();
#endif
	sig2.GenerateImage("shapes2.png");
	textures2.GenerateImage("textures2.png");
#ifdef _OPENMP
    const double image2ShapesAndTexturesEnd = omp_get_wtime();
	cout<<"("<<(image2ShapesAndTexturesEnd)-(image2ShapesAndTexturesStart)<<"s)"<<endl;
#endif

	// Put all descriptors together on a single vector, as we did for the first image
	vector<std::pair<int, double*> > descI2Shapes = vector<std::pair<int, double*> >(qtyShapes);
	vector<std::pair<int, double*> > descI2Textures = vector<std::pair<int, double*> >(qtyTextures);
	qtyShapes = 0;
	qtyTextures = 0;
	for (unsigned int i=0;i<descTempShapes.size();i++)
	{
		for (unsigned int j=0;j<descTempShapes[i].second.size();j++)
		{
			descI2Shapes[qtyShapes].first = descTempShapes[i].first;
			descI2Shapes[qtyShapes++].second = descTempShapes[i].second.at(j);
		}
	}
	for (unsigned int i=0;i<descTempTextures.size();i++)
	{
		for (unsigned int j=0;j<descTempTextures[i].second.size();j++)
		{
			descI2Textures[qtyTextures].first = descTempTextures[i].first;
			descI2Textures[qtyTextures++].second = descTempTextures[i].second.at(j);
		}
	}

	
	// Now try to match shape descriptors

#ifdef _OPENMP
	double matchTime = omp_get_wtime();
#endif	
	vector<std::pair<CMSER*,CMSER*> > shapeMatches = vector<std::pair<CMSER*,CMSER*> >(descI1Shapes.size());
	std::vector<std::pair<int,int> > mserShapeMatched = std::vector<std::pair<int,int> >(descI2Shapes.size());
	CDescriptorMatcher dm = CDescriptorMatcher(descI1Shapes, descI2Shapes, listMSER1, listMSER2);
	dm.MatchDescriptors(shapeMatches, mserShapeMatched, CParams::Instance()->matchThreshold );

	// Uncomment this code if generating an image of the matches MSERs is required
	//for (unsigned int i=0;i<shapeMatches.size();i++)
	//{
	//	cout << "Shape Match: (" << shapeMatches[i].first->GetCentroid()->first << "," <<  shapeMatches[i].first->GetCentroid()->second << ") with (" << shapeMatches[i].second->GetCentroid()->first << "," << shapeMatches[i].second->GetCentroid()->second << ")" << endl;
	//	float *fig1 = sig.getShape(mserShapeMatched[i].first);
	//	float *fig2 = sig2.getShape(mserShapeMatched[i].second);
	//	CShapeImageGenerator match = CShapeImageGenerator(2, false);
	//	match.addShape(fig1, 0);
	//	match.addShape(fig2, 1);
	//	char filename[256];
	//	sprintf(filename, "match%d.png", i);
	//	match.GenerateImage(filename);
	//}

    
	// Generate an image with no matches (to show in webpage)
	vector<std::pair<CMSER*,CMSER*> > noMatches = vector<std::pair<CMSER*,CMSER*> >(0);
	CMatchesImageGenerator migTemp = CMatchesImageGenerator(*img1, *img2);
	migTemp.GenerateImage("output.png", noMatches);

	// Generate an image to show shape matches
	CMatchesImageGenerator mig = CMatchesImageGenerator(*img1, *img2);
	mig.GenerateImage("outputShape.png", shapeMatches);
    
    
	// Try to match texture descriptors
	vector<std::pair<CMSER*,CMSER*> > textureMatches = vector<std::pair<CMSER*,CMSER*> >(descI1Textures.size());
	std::vector<std::pair<int,int> > mserTextureMatched = std::vector<std::pair<int,int> >(descI2Textures.size());

	CDescriptorMatcher dm2 = CDescriptorMatcher(descI1Textures, descI2Textures, listMSER1, listMSER2);
	dm2.MatchDescriptors(textureMatches, mserTextureMatched, CParams::Instance()->matchThreshold );

	// Uncomment this code if generating an image of the matches MSERs is required
	//for (int i=0;i<textureMatches.size();i++)
	//{
		//cout << "Texture Match: (" << textureMatches[i].first->GetCentroid()->first << "," <<  textureMatches[i].first->GetCentroid()->second << ") with (" << textureMatches[i].second->GetCentroid()->first << "," << textureMatches[i].second->GetCentroid()->second << ")" << endl;
		//unsigned char *fig1 = textures.getTexture(mserTextureMatched[i].first);
		//unsigned char *fig2 = textures2.getTexture(mserTextureMatched[i].second);
		//CTextureImageGenerator match = CTextureImageGenerator(2, false);
		//match.addTexture(fig1, 0);
		//match.addTexture(fig2, 1);
		//char filename[256];
		//sprintf(filename, "cdsqklsq.png", i);
		//match.GenerateImage(filename);
	//}
 
	// Generate an image to show texture matches
	CMatchesImageGenerator mig2 = CMatchesImageGenerator(*img1, *img2);
	mig2.GenerateImage("outputTexture.png", textureMatches);
#ifdef _OPENMP
    printf("Matching time = %lf seconds\n", (omp_get_wtime() - matchTime));
	double ransacTime = omp_get_wtime();
#endif
	// Now apply RANSAC to filter bad matches for both shape and texture matches
	cout<<"For shape, ";
	CRansacWrapper ransac = CRansacWrapper(CParams::Instance()->ransacMaxErr);
	vector<std::pair<CMSER*,CMSER*> > filteredShapeMatchesRansac = vector<std::pair<CMSER*,CMSER*> >(shapeMatches.size());
	ransac.FilterResults(shapeMatches, filteredShapeMatchesRansac);
	if (filteredShapeMatchesRansac.size() > 0)
	{
		CMatchesImageGenerator mig3 = CMatchesImageGenerator(*img1, *img2);
		mig3.GenerateImage("outputShapeRANSAC.png", filteredShapeMatchesRansac);
		//for (int j=0;j<filteredShapeMatchesRansac.size();j++)
		//{
		//	float *fig1 = filteredShapeMatchesRansac[j].first->GetBinaryMaskImage()->GetFloatVersion();
		//	float *fig2 = filteredShapeMatchesRansac[j].second->GetBinaryMaskImage()->GetFloatVersion();
		//	//float *fig1 = sig.getShape(filteredShapeMatchesRansac[i].first);
		//	//float *fig2 = sig2.getShape(filteredShapeMatchesRansac[i].second);
		//	CShapeImageGenerator match = CShapeImageGenerator(2, false);
		//	match.addShape(fig1, 0);
		//	match.addShape(fig2, 1);
		//	char filename[256];
		//	sprintf(filename, "c://matchRansac%d.png", j);
		//	match.GenerateImage(filename);

		//}

	}
	else
	{
		CMatchesImageGenerator mig3 = CMatchesImageGenerator(*img1, *img2);
		mig3.GenerateImage("outputShapeRANSAC.png", shapeMatches);
		cout << "No RANSAC filtering due to small amount of initial matches. Showing all." << endl;
	}
#ifdef _OPENMP
    printf("Shape RANSAC time = %lf seconds\n", (omp_get_wtime() - ransacTime));
	ransacTime = omp_get_wtime();
#endif

	cout<<"For texture, ";
	CRansacWrapper ransac2 = CRansacWrapper();
	vector<std::pair<CMSER*,CMSER*> > filteredTextureMatchesRansac = vector<std::pair<CMSER*,CMSER*> >(textureMatches.size());
	ransac2.FilterResults(textureMatches, filteredTextureMatchesRansac);
	if (filteredTextureMatchesRansac.size() > 0)
	{
		CMatchesImageGenerator mig = CMatchesImageGenerator(*img1, *img2);
		mig.GenerateImage("outputTextureRANSAC.png", filteredTextureMatchesRansac);
	}
	else
	{
		CMatchesImageGenerator mig = CMatchesImageGenerator(*img1, *img2);
		mig.GenerateImage("outputTextureRANSAC.png", textureMatches);
		cout << "No RANSAC filtering due to small amount of initial matches. Showing all." << endl;
	}
#ifdef _OPENMP
    printf("Texture RANSAC time = %lf seconds\n", (omp_get_wtime() - ransacTime));
#endif
	
	// Create Javascript file for webpage MSER displaying
	
	//createJSFile("test.js",(filteredShapeMatchesRansac.size() > 0 ? filteredShapeMatchesRansac: shapeMatches) , (filteredTextureMatchesRansac.size() > 0 ? filteredTextureMatchesRansac : textureMatches) , shapeMatches , textureMatches, img1->Width()+img2->Width(),(img1->Height()>img2->Height())?img1->Height():img2->Height(),img1->Width());
	//createJSFile("test.js",(filteredShapeMatchesRansac.size() > 0 ? filteredShapeMatchesRansac: shapeMatches) , (filteredTextureMatchesRansac.size() > 0 ? filteredTextureMatchesRansac : textureMatches) , img1->Width()+img2->Width(),(img1->Height()>img2->Height())?img1->Height():img2->Height(),img1->Width());
	createJSFile("test.js",(filteredShapeMatchesRansac.size() > 0 ? filteredShapeMatchesRansac: shapeMatches) , (filteredTextureMatchesRansac.size() > 0 ? filteredTextureMatchesRansac : textureMatches) , img1->Width());
		

	// Free memory
#ifdef _OPENMP
	double freeMemTime = omp_get_wtime();
#endif	
	#pragma omp parallel for ordered schedule(dynamic)
	for (int i=0;i< (int) listMSER1->size();i++) 	
	{ 
		delete listMSER1->at(i); 
	}
	#pragma omp parallel for ordered schedule(dynamic)
	for (int i=0;i< (int) listMSER2->size();i++) 
	{ 
		delete listMSER2->at(i); 
	}
	#pragma omp parallel for ordered schedule(dynamic)
	for (int i=0;i<(int) descI1Shapes.size();i++)
	{ 
		//cout << "Freeing memory " << i << endl;
		delete[] descI1Shapes.at(i).second; 
	}
	#pragma omp parallel for ordered schedule(dynamic)
	for (int i=0;i<(int) descI1Textures.size();i++)
	{ 
		//cout << "Freeing memory " << i << endl;
		delete[] descI1Textures.at(i).second; 
	}
	#pragma omp parallel for ordered schedule(dynamic)
	for (int i=0;i<(int) descI2Shapes.size();i++) { 
		//cout << "Freeing memory " << i << endl;
		delete[] descI2Shapes.at(i).second; 
	}
	#pragma omp parallel for ordered schedule(dynamic)
	for (int i=0;i<(int) descI2Textures.size();i++) { 
		//cout << "Freeing memory " << i << endl;
		delete[] descI2Textures.at(i).second; 
	}
	//cout << "Freeing memory2" << endl;

	delete listMSER1;
	delete listMSER2;
	delete img1;
	delete img2;

#ifdef _OPENMP
    cout << "Freeing memory time = " << (omp_get_wtime() - freeMemTime) << "s" << endl;

    cout << "Total time " << (omp_get_wtime() - mserStart) << "s" << endl;
	cout<<"Computation time, MSER: "<<-mserStart+mserEnd<<"s"<<endl;
	cout<<"Computation time, shape descriptors: "<<(shape1End+shape2End)-(shape1Start+shape2Start)<<"s"<<endl;
	cout<<"Computation time, texture descriptors: "<<(texture1End+texture2End)-(texture1Start+texture2Start)<<"s"<<endl;
#endif
	CParams::KillInstance();
	return EXIT_SUCCESS;
}

/**
 * @fn	CMserPlusSiftMain::CMserPlusSiftMain(void)
 *
 * @brief	Default constructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CMserPlusSiftMain::CMserPlusSiftMain(void)
{
}

/**
 * @fn	CMserPlusSiftMain::~CMserPlusSiftMain(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	18/02/2014
 */

CMserPlusSiftMain::~CMserPlusSiftMain(void)
{
}
