﻿/**
 * @file	C:\Users\Tato\Documents\GitHub\IPOL_MSER\MSERSIFT\src\MultiscaleMSER.cpp
 *
 * @brief	Implements the multiscale mser class.
 */

#include "MultiscaleMSER.h"
#include "GaussianBlur.h"
#include "ImageResampler.h"
#include "HelperMethods.h"
#include "IO_PNG/io_png.h"
#include "Params.h"


/**
 * @fn	CMultiscaleMSER::CMultiscaleMSER(CGrayscaleImage *img)
 *
 * @brief	Constructor.
 *
 * @author	Tato
 * @date	21/02/2014
 *
 * @param [in]	img		The image.
 */

CMultiscaleMSER::CMultiscaleMSER(CGrayscaleImage *img, int scales)
{
	_images = new std::vector<CGrayscaleImage*>(scales);
	_images->at(0) = new CGrayscaleImage(img);
	_scales = scales;

}


/**
 * @fn	const CMSERList *CMultiscaleMSER::GetMSERList()
 *
 * @brief	Gets mser list.
 *
 * @author	Tato
 * @date	21/02/2014
 *
 * @return	null if it fails, else the mser list.
 */

const CMSERList *CMultiscaleMSER::GetMSERList()
{
	// First take the MSERs from the input image (finer scale)
	CMSERWrapper *mser = new CMSERWrapper(*(_images->at(0)), 0);
	//unsigned char *saveImg = _img->GetGrayscaleVersion();
	//write_png_u8("c:\\test.png", saveImg, img->Width(), img->Height(), 1);


	vector<CMSERList *> list = vector<CMSERList *>(_scales);
	list[0] = mser->GetList();
	delete mser;
	int totalMSER = (int) list[0]->size();

	// Now iterate through scales by applying Gaussian and subsampling
	for (int i = 1; i<_scales; i++)
	{
		CGaussianBlur gb;
		// Apply 6-tap Gaussian blur with sigma=1
		_images->at(i) = new CGrayscaleImage(_images->at(i - 1)->Width(), _images->at(i - 1)->Height());
		gb.ApplyOptimizedGaussianBlur(*_images->at(i - 1), _images->at(i), 6, 1.0);
		CImageResampler imRes = CImageResampler(_images->at(i));
		CGrayscaleImage *res = imRes.SubsampleImageInHalf();
		CMSERWrapper *mser = new CMSERWrapper(*res, i);
		list[i] = mser->GetList();
		totalMSER += (int) list[i]->size();
		delete mser;
		delete _images->at(i);
		_images->at(i) = res;
	}
	//for (int i=0;i<scales;i++)
	//{
	//	unsigned char *aux = _images->at(i)->GetGrayscaleVersion();
	//	char caca[200];
	//	sprintf(caca, "C:\\scale%d.png",i);
	//	write_png_u8(caca, aux, _images->at(i)->Width(), _images->at(i)->Height(),1);
	//	delete aux;

	//}
	cout << "Found " << totalMSER << " MSERs and";	//cout << "Removing duplicates..." << endl;
	int amountRemoved = 0;
	for (int i = 0; i<_scales - 1; i++)
	{
		CMSERList *finer = list[i];
		CMSERList *coarser = list[i + 1];
		// Delete if centroid 
		for (unsigned int it1 = 0; it1 < finer->size(); it1++)
		{
			CMSER *mserFine = finer->at(it1);
			//const CPixel *centroidFine = mserFine->GetCentroid();
			for (unsigned int it2 = 0; it2 < coarser->size(); it2++)
			{

				CMSER *mserCoarse = coarser->at(it2);
				float ellipseLen = coarser->at(it2)->GetMinorEllipseAxisLength();
				if (ellipseLen < CParams::Instance()->minimumEllipseLength)
				{
					delete mserCoarse;
					coarser->erase(coarser->begin() + it2);
					amountRemoved++;
					it2--;
					continue;
				}
				// The centroid distance should be smaller than 4 pixels in the finer grid
				if (mserFine->HasSameCentroid(*mserCoarse))
				{
					// For two areas, a1 and a2, we additionally require that abs(a1 − a2)/max(a1, a2) < 0.2.
					if (mserFine->HasSimilarArea(*mserCoarse))
					{
						// Remove the finer MSER
						delete mserFine;
						finer->erase(finer->begin() + it1);
						it1--;
						amountRemoved++;
						break;
					}
				}
			}
		}
	}
	int MSERQty = totalMSER - amountRemoved;
	cout << ", removed " << 100.0 * ((double)amountRemoved / (double)totalMSER) << "%. Left: " << MSERQty << " MSERs" << endl;
	CMSERList *resList = new CMSERList(MSERQty);
	int elem = 0;
	for (int i = 0; i<_scales; i++)
	{
		for (unsigned int j = 0; j<list[i]->size(); j++)
		{
			resList->at(elem) = list[i]->at(j);
			resList->at(elem)->SetId(elem);
			elem++;
		}
	}
	for (int i = 0; i<_scales; i++)
	{
		delete list[i];
	}
	return resList;
}

/**
 * @fn	CMultiscaleMSER::~CMultiscaleMSER(void)
 *
 * @brief	Destructor.
 *
 * @author	Tato
 * @date	21/02/2014
 */

CMultiscaleMSER::~CMultiscaleMSER(void)
{
	for  (unsigned int i=0;i<_images->size();i++)
		delete _images->at(i);
	delete _images;
	//delete _list;
}
